# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.status = skip

from scapy.fields import BitField, FlagsField, ScalingField
from scapy.contrib.automotive.obd.packet import OBD_Packet


# See https://en.wikipedia.org/wiki/OBD-II_PIDs for further information
# PID = Parameter IDentification


class OBD_PID60(OBD_Packet):
    name = "PID_60_PIDsSupported"
    fields_desc = [
        FlagsField('supported_pids', 0, 32, [
            'PID80',
            'PID7F',
            'PID7E',
            'PID7D',
            'PID7C',
            'PID7B',
            'PID7A',
            'PID79',
            'PID78',
            'PID77',
            'PID76',
            'PID75',
            'PID74',
            'PID73',
            'PID72',
            'PID71',
            'PID70',
            'PID6F',
            'PID6E',
            'PID6D',
            'PID6C',
            'PID6B',
            'PID6A',
            'PID69',
            'PID68',
            'PID67',
            'PID66',
            'PID65',
            'PID64',
            'PID63',
            'PID62',
            'PID61'
        ])
    ]


class OBD_PID61(OBD_Packet):
    name = "PID_61_DriverSDemandEnginePercentTorque"
    fields_desc = [
        ScalingField('data', 0, unit="%", offset=-125.0)
    ]


class OBD_PID62(OBD_Packet):
    name = "PID_62_ActualEnginePercentTorque"
    fields_desc = [
        ScalingField('data', 0, unit="%", offset=-125.0)
    ]


class OBD_PID63(OBD_Packet):
    name = "PID_63_EngineReferenceTorque"
    fields_desc = [
        ScalingField('data', 0, unit="Nm", fmt="H")
    ]


class OBD_PID64(OBD_Packet):
    name = "PID_64_EnginePercentTorqueData"
    fields_desc = [
        ScalingField('at_point1', 0, unit="%", offset=-125.0),
        ScalingField('at_point2', 0, unit="%", offset=-125.0),
        ScalingField('at_point3', 0, unit="%", offset=-125.0),
        ScalingField('at_point4', 0, unit="%", offset=-125.0),
        ScalingField('at_point5', 0, unit="%", offset=-125.0)
    ]


class OBD_PID65(OBD_Packet):
    name = "PID_65_AuxiliaryInputOutputSupported"
    fields_desc = [
        BitField('reserved1', 0, 4),
        BitField('glow_plug_lamp_status_supported', 0, 1),
        BitField('manual_trans_neutral_drive_status_supported', 0, 1),
        BitField('auto_trans_neutral_drive_status_supported', 0, 1),
        BitField('power_take_off_status_supported', 0, 1),

        BitField('reserved2', 0, 4),
        BitField('glow_plug_lamp_status', 0, 1),
        BitField('manual_trans_neutral_drive_status', 0, 1),
        BitField('auto_trans_neutral_drive_status', 0, 1),
        BitField('power_take_off_status', 0, 1),
    ]


class OBD_PID66(OBD_Packet):
    name = "PID_66_MassAirFlowSensor"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('sensor_b_supported', 0, 1),
        BitField('sensor_a_supported', 0, 1),
        ScalingField('sensor_a', 0, scaling=0.03125, unit="g/s", fmt="H"),
        ScalingField('sensor_b', 0, scaling=0.03125, unit="g/s", fmt="H"),
    ]


class OBD_PID67(OBD_Packet):
    name = "PID_67_EngineCoolantTemperature"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('sensor2_supported', 0, 1),
        BitField('sensor1_supported', 0, 1),
        ScalingField('sensor1', 0, unit="deg. C", offset=-40.0),
        ScalingField('sensor2', 0, unit="deg. C", offset=-40.0)
    ]


class OBD_PID68(OBD_Packet):
    name = "PID_68_IntakeAirTemperatureSensor"
    fields_desc = [
        BitField('reserved', 0, 2),
        BitField('bank2_sensor3_supported', 0, 1),
        BitField('bank2_sensor2_supported', 0, 1),
        BitField('bank2_sensor1_supported', 0, 1),
        BitField('bank1_sensor3_supported', 0, 1),
        BitField('bank1_sensor2_supported', 0, 1),
        BitField('bank1_sensor1_supported', 0, 1),
        ScalingField('bank1_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank1_sensor2', 0, unit="deg. C", offset=-40),
        ScalingField('bank1_sensor3', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor2', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor3', 0, unit="deg. C", offset=-40)
    ]


class OBD_PID69(OBD_Packet):
    name = "PID_69_CommandedEgrAndEgrError"
    fields_desc = [
        BitField('reserved', 0, 2),
        BitField('egr_b_error_supported', 0, 1),
        BitField('actual_egr_b_duty_cycle_supported', 0, 1),
        BitField('commanded_egr_b_duty_cycle_supported', 0, 1),
        BitField('egr_a_error_supported', 0, 1),
        BitField('actual_egr_a_duty_cycle_supported', 0, 1),
        BitField('commanded_egr_a_duty_cycle_supported', 0, 1),
        ScalingField('commanded_egr_a_duty_cycle', 0, scaling=100 / 255.,
                     unit="%"),
        ScalingField('actual_egr_a_duty_cycle', 0, scaling=100 / 255.,
                     unit="%"),
        ScalingField('egr_a_error', 0, scaling=100 / 128., unit="%",
                     offset=-100),
        ScalingField('commanded_egr_b_duty_cycle', 0, scaling=100 / 255.,
                     unit="%"),
        ScalingField('actual_egr_b_duty_cycle', 0, scaling=100 / 255.,
                     unit="%"),
        ScalingField('egr_b_error', 0, scaling=100 / 128., unit="%",
                     offset=-100),
    ]


class OBD_PID6A(OBD_Packet):
    name = "PID_6A_CommandedDieselIntakeAirFlowControl" \
           "AndRelativeIntakeAirFlowPosition"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('relative_intake_air_flow_b_position_supported', 0, 1),
        BitField('commanded_intake_air_flow_b_control_supported', 0, 1),
        BitField('relative_intake_air_flow_a_position_supported', 0, 1),
        BitField('commanded_intake_air_flow_a_control_supported', 0, 1),
        ScalingField('commanded_intake_air_flow_a_control', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('relative_intake_air_flow_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('commanded_intake_air_flow_b_control', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('relative_intake_air_flow_b_position', 0,
                     scaling=100 / 255., unit="%"),
    ]


class OBD_PID6B(OBD_Packet):
    name = "PID_6B_ExhaustGasRecirculationTemperature"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('bank2_sensor2_supported', 0, 1),
        BitField('bank2_sensor1_supported', 0, 1),
        BitField('bank1_sensor2_supported', 0, 1),
        BitField('bank1_sensor1_supported', 0, 1),
        ScalingField('bank1_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank1_sensor2', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor2', 0, unit="deg. C", offset=-40),
    ]


class OBD_PID6C(OBD_Packet):
    name = "PID_6C_CommandedThrottleActuatorControlAndRelativeThrottlePosition"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('relative_throttle_b_position_supported', 0, 1),
        BitField('commanded_throttle_actuator_b_control_supported', 0, 1),
        BitField('relative_throttle_a_position_supported', 0, 1),
        BitField('commanded_throttle_actuator_a_control_supported', 0, 1),
        ScalingField('commanded_throttle_actuator_a_control', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('relative_throttle_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('commanded_throttle_actuator_b_control', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('relative_throttle_b_position', 0,
                     scaling=100 / 255., unit="%"),
    ]


class OBD_PID6D(OBD_Packet):
    name = "PID_6D_FuelPressureControlSystem"
    fields_desc = [
        BitField('reserved', 0, 5),
        BitField('fuel_temperature_supported', 0, 1),
        BitField('fuel_rail_pressure_supported', 0, 1),
        BitField('commanded_fuel_rail_pressure_supported', 0, 1),
        ScalingField('commanded_fuel_rail_pressure', 0, scaling=10, unit="kPa",
                     fmt='H'),
        ScalingField('fuel_rail_pressure', 0, scaling=10, unit="kPa",
                     fmt='H'),
        ScalingField('fuel_rail_temperature', 0, unit="deg. C", offset=-40)
    ]


class OBD_PID6E(OBD_Packet):
    name = "PID_6E_InjectionPressureControlSystem"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('injection_control_pressure_supported', 0, 1),
        BitField('commanded_injection_control_pressure_supported', 0, 1),
        ScalingField('commanded_injection_control_pressure', 0, scaling=10,
                     unit="kPa", fmt='H'),
        ScalingField('injection_control_pressure', 0, scaling=10,
                     unit="kPa", fmt='H'),
    ]


class OBD_PID6F(OBD_Packet):
    name = "PID_6F_TurbochargerCompressorInletPressure"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('sensor_b_supported', 0, 1),
        BitField('sensor_a_supported', 0, 1),
        ScalingField('sensor_a', 0, unit="kPa"),
        ScalingField('sensor_b', 0, unit="kPa"),
    ]


class OBD_PID70(OBD_Packet):
    name = "PID_70_BoostPressureControl"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('boost_pressure_sensor_b_supported', 0, 1),
        BitField('commanded_boost_pressure_b_supported', 0, 1),
        BitField('boost_pressure_sensor_a_supported', 0, 1),
        BitField('commanded_boost_pressure_a_supported', 0, 1),
        ScalingField('commanded_boost_pressure_a', 0, scaling=0.03125,
                     unit="kPa", fmt='H'),
        ScalingField('boost_pressure_sensor_a', 0, scaling=0.03125,
                     unit="kPa", fmt='H'),
        ScalingField('commanded_boost_pressure_b', 0, scaling=0.03125,
                     unit="kPa", fmt='H'),
        ScalingField('boost_pressure_sensor_b', 0, scaling=0.03125,
                     unit="kPa", fmt='H'),
    ]


class OBD_PID71(OBD_Packet):
    name = "PID_71_VariableGeometryTurboControl"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('vgt_b_position_supported', 0, 1),
        BitField('commanded_vgt_b_position_supported', 0, 1),
        BitField('vgt_a_position_supported', 0, 1),
        BitField('commanded_vgt_a_position_supported', 0, 1),
        ScalingField('commanded_variable_geometry_turbo_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('variable_geometry_turbo_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('commanded_variable_geometry_turbo_b_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('variable_geometry_turbo_b_position', 0,
                     scaling=100 / 255., unit="%"),
    ]


class OBD_PID72(OBD_Packet):
    name = "PID_72_WastegateControl"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('wastegate_b_position_supported', 0, 1),
        BitField('commanded_wastegate_b_position_supported', 0, 1),
        BitField('wastegate_a_position_supported', 0, 1),
        BitField('commanded_wastegate_a_position_supported', 0, 1),
        ScalingField('commanded_wastegate_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('wastegate_a_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('commanded_wastegate_b_position', 0,
                     scaling=100 / 255., unit="%"),
        ScalingField('wastegate_b_position', 0,
                     scaling=100 / 255., unit="%"),
    ]


class OBD_PID73(OBD_Packet):
    name = "PID_73_ExhaustPressure"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('sensor_bank2_supported', 0, 1),
        BitField('sensor_bank1_supported', 0, 1),
        ScalingField('sensor_bank1', 0, scaling=0.01, unit="kPa", fmt='H'),
        ScalingField('sensor_bank2', 0, scaling=0.01, unit="kPa", fmt='H'),
    ]


class OBD_PID74(OBD_Packet):
    name = "PID_74_TurbochargerRpm"
    fields_desc = [
        BitField('reserved', 0, 6),
        BitField('b_supported', 0, 1),
        BitField('a_supported', 0, 1),
        ScalingField('a_rpm', 0, unit="min-1", fmt='H'),
        ScalingField('b_rpm', 0, unit="min-1", fmt='H'),
    ]


class OBD_PID75(OBD_Packet):
    name = "PID_75_TurbochargerATemperature"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('turbo_a_turbine_outlet_temperature_supported', 0, 1),
        BitField('turbo_a_turbine_inlet_temperature_supported', 0, 1),
        BitField('turbo_a_compressor_outlet_temperature_supported', 0, 1),
        BitField('turbo_a_compressor_inlet_temperature_supported', 0, 1),
        ScalingField('turbocharger_a_compressor_inlet_temperature', 0,
                     unit="deg. C", offset=-40),
        ScalingField('turbocharger_a_compressor_outlet_temperature', 0,
                     unit="deg. C", offset=-40),
        ScalingField('turbocharger_a_turbine_inlet_temperature', 0,
                     unit="deg. C", offset=-40, fmt='H',
                     scaling=0.1),
        ScalingField('turbocharger_a_turbine_outlet_temperature', 0,
                     unit="deg. C", offset=-40, fmt='H',
                     scaling=0.1),
    ]


class OBD_PID76(OBD_Packet):
    name = "PID_76_TurbochargerBTemperature"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('turbo_a_turbine_outlet_temperature_supported', 0, 1),
        BitField('turbo_a_turbine_inlet_temperature_supported', 0, 1),
        BitField('turbo_a_compressor_outlet_temperature_supported', 0, 1),
        BitField('turbo_a_compressor_inlet_temperature_supported', 0, 1),
        ScalingField('turbocharger_a_compressor_inlet_temperature', 0,
                     unit="deg. C", offset=-40),
        ScalingField('turbocharger_a_compressor_outlet_temperature', 0,
                     unit="deg. C", offset=-40),
        ScalingField('turbocharger_a_turbine_inlet_temperature', 0,
                     unit="deg. C", offset=-40, fmt='H',
                     scaling=0.1),
        ScalingField('turbocharger_a_turbine_outlet_temperature', 0,
                     unit="deg. C", offset=-40, fmt='H',
                     scaling=0.1),
    ]


class OBD_PID77(OBD_Packet):
    name = "PID_77_ChargeAirCoolerTemperature"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('bank2_sensor2_supported', 0, 1),
        BitField('bank2_sensor1_supported', 0, 1),
        BitField('bank1_sensor2_supported', 0, 1),
        BitField('bank1_sensor1_supported', 0, 1),
        ScalingField('bank1_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank1_sensor2', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor1', 0, unit="deg. C", offset=-40),
        ScalingField('bank2_sensor2', 0, unit="deg. C", offset=-40),
    ]


class _OBD_PID_ExhaustGasTemperatureBank(OBD_Packet):
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('sensor4_supported', 0, 1),
        BitField('sensor3_supported', 0, 1),
        BitField('sensor2_supported', 0, 1),
        BitField('sensor1_supported', 0, 1),
        ScalingField('sensor1', 0, unit="deg. C", offset=-40,
                     scaling=0.1, fmt='H'),
        ScalingField('sensor2', 0, unit="deg. C", offset=-40,
                     scaling=0.1, fmt='H'),
        ScalingField('sensor3', 0, unit="deg. C", offset=-40,
                     scaling=0.1, fmt='H'),
        ScalingField('sensor4', 0, unit="deg. C", offset=-40,
                     scaling=0.1, fmt='H'),
    ]


class OBD_PID78(_OBD_PID_ExhaustGasTemperatureBank):
    name = "PID_78_ExhaustGasTemperatureBank1"


class OBD_PID79(_OBD_PID_ExhaustGasTemperatureBank):
    name = "PID_79_ExhaustGasTemperatureBank2"


class _OBD_PID_DieselParticulateFilter(OBD_Packet):
    fields_desc = [
        BitField('reserved', 0, 5),
        BitField('outlet_pressure_supported', 0, 1),
        BitField('inlet_pressure_supported', 0, 1),
        BitField('delta_pressure_supported', 0, 1),
        ScalingField('delta_pressure', 0,
                     unit='kPa', offset=-327.68, scaling=0.01, fmt='H'),
        ScalingField('particulate_filter', 0,
                     unit='kPa', scaling=0.01, fmt='H'),
        ScalingField('outlet_pressure', 0,
                     unit='kPa', scaling=0.01, fmt='H'),
    ]


class OBD_PID7A(_OBD_PID_DieselParticulateFilter):
    name = "PID_7A_DieselParticulateFilter1"


class OBD_PID7B(_OBD_PID_DieselParticulateFilter):
    name = "PID_7B_DieselParticulateFilter2"


class OBD_PID7C(OBD_Packet):
    name = "PID_7C_DieselParticulateFilterTemperature"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('bank2_outlet_temperature_supported', 0, 1),
        BitField('bank2_inlet_temperature_supported', 0, 1),
        BitField('bank1_outlet_temperature_supported', 0, 1),
        BitField('bank1_inlet_temperature_supported', 0, 1),
        ScalingField('bank1_inlet_temperature_sensor', 0,
                     unit="deg. C", offset=-40, scaling=0.1, fmt='H'),
        ScalingField('bank1_outlet_temperature_sensor', 0,
                     unit="deg. C", offset=-40, scaling=0.1, fmt='H'),
        ScalingField('bank2_inlet_temperature_sensor', 0,
                     unit="deg. C", offset=-40, scaling=0.1, fmt='H'),
        ScalingField('bank2_outlet_temperature_sensor', 0,
                     unit="deg. C", offset=-40, scaling=0.1, fmt='H'),
    ]


class OBD_PID7D(OBD_Packet):
    name = "PID_7D_NoxNteControlAreaStatus"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('nte_deficiency_for_nox_active_area', 0, 1),
        BitField('inside_manufacturer_specific_nox_nte_carve_out_area', 0, 1),
        BitField('outside', 0, 1),
        BitField('inside', 0, 1),
    ]


class OBD_PID7E(OBD_Packet):
    name = "PID_7E_PmNteControlAreaStatus"
    fields_desc = [
        BitField('reserved', 0, 4),
        BitField('nte_deficiency_for_pm_active_area', 0, 1),
        BitField('inside_manufacturer_specific_pm_nte_carve_out_area', 0, 1),
        BitField('outside', 0, 1),
        BitField('inside', 0, 1),
    ]


class OBD_PID7F(OBD_Packet):
    name = "PID_7F_EngineRunTime"
    fields_desc = [
        BitField('reserved', 0, 5),
        BitField('total_with_pto_active_supported', 0, 1),
        BitField('total_idle_supported', 0, 1),
        BitField('total_supported', 0, 1),
        ScalingField('total', 0, unit='sec', fmt='Q'),
        ScalingField('total_idle', 0, unit='sec', fmt='Q'),
        ScalingField('total_with_pto_active', 0, unit='sec', fmt='Q'),
    ]



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>
# Copyright (C) Enrico Pozzobon <enricopozzobon@gmail.com>
# Copyright (C) Alexander Schroeder <alexander1.schroeder@st.othr.de>

# scapy.contrib.description = ISO-TP (ISO 15765-2) Utilities
# scapy.contrib.status = library

import struct

from scapy.config import conf
from scapy.utils import EDecimal
from scapy.packet import Packet
from scapy.sessions import DefaultSession
from scapy.supersocket import SuperSocket
from scapy.contrib.isotp.isotp_packet import ISOTP, N_PCI_CF, N_PCI_SF, \
    N_PCI_FF, N_PCI_FC

# Typing imports
from typing import (
    cast,
    Iterable,
    Iterator,
    Optional,
    Union,
    List,
    Tuple,
    Dict,
    Any,
    Type,
)


class ISOTPMessageBuilderIter(object):
    """
    Iterator class for ISOTPMessageBuilder
    """
    slots = ["builder"]

    def __init__(self, builder):
        # type: (ISOTPMessageBuilder) -> None
        self.builder = builder

    def __iter__(self):
        # type: () -> ISOTPMessageBuilderIter
        return self

    def __next__(self):
        # type: () -> ISOTP
        while self.builder.count:
            p = self.builder.pop()
            if p is None:
                break
            else:
                return p
        raise StopIteration

    next = __next__


class ISOTPMessageBuilder(object):
    """
    Initialize a ISOTPMessageBuilder object

    Utility class to build ISOTP messages out of CAN frames, used by both
    ISOTP.defragment() and ISOTPSession.

    This class attempts to interpret some CAN frames as ISOTP frames, both with
    and without extended addressing at the same time. For example, if an
    extended address of 07 is being used, all frames will also be interpreted
    as ISOTP single-frame messages.

    CAN frames are fed to an ISOTPMessageBuilder object with the feed() method
    and the resulting ISOTP frames can be extracted using the pop() method.

    :param use_ext_address: True for only attempting to defragment with
                         extended addressing, False for only attempting
                         to defragment without extended addressing,
                         or None for both
    :param rx_id: Destination Identifier
    :param basecls: The class of packets that will be returned,
                    defaults to ISOTP
    """

    class Bucket(object):
        """
        Helper class to store not finished ISOTP messages while building.
        """

        def __init__(self, total_len, first_piece, ts):
            # type: (int, bytes, Union[EDecimal, float]) -> None
            self.pieces = list()  # type: List[bytes]
            self.total_len = total_len
            self.current_len = 0
            self.ready = None  # type: Optional[bytes]
            self.tx_id = None  # type: Optional[int]
            self.ext_address = None  # type: Optional[int]
            self.time = ts  # type: Union[float, EDecimal]
            self.push(first_piece)

        def push(self, piece):
            # type: (bytes) -> None
            self.pieces.append(piece)
            self.current_len += len(piece)
            if self.current_len >= self.total_len:
                isotp_data = b"".join(self.pieces)
                self.ready = isotp_data[:self.total_len]

    def __init__(
            self,
            use_ext_address=None,  # type: Optional[bool]
            rx_id=None,  # type: Optional[Union[int, List[int], Iterable[int]]]
            basecls=ISOTP  # type: Type[ISOTP]
    ):
        # type: (...) -> None
        self.ready = []  # type: List[Tuple[int, Optional[int], ISOTPMessageBuilder.Bucket]]  # noqa: E501
        self.buckets = {}  # type: Dict[Tuple[Optional[int], int, int], ISOTPMessageBuilder.Bucket]  # noqa: E501
        self.use_ext_addr = use_ext_address
        self.basecls = basecls
        self.rx_ids = None  # type: Optional[Iterable[int]]
        self.last_ff = None  # type: Optional[Tuple[Optional[int], int, int]]
        self.last_ff_ex = None  # type: Optional[Tuple[Optional[int], int, int]]  # noqa: E501
        if rx_id is not None:
            if isinstance(rx_id, list):
                self.rx_ids = rx_id
            elif isinstance(rx_id, int):
                self.rx_ids = [rx_id]
            elif hasattr(rx_id, "__iter__"):
                self.rx_ids = rx_id
            else:
                raise TypeError("Invalid type for argument rx_id!")

    def feed(self, can):
        # type: (Union[Iterable[Packet], Packet]) -> None
        """Attempt to feed an incoming CAN frame into the state machine"""
        if not isinstance(can, Packet) and hasattr(can, "__iter__"):
            for p in can:
                self.feed(p)
            return

        if not isinstance(can, Packet):
            return

        if self.rx_ids is not None and can.identifier not in self.rx_ids:
            return

        data = bytes(can.data)

        if len(data) > 1 and self.use_ext_addr is not True:
            self._try_feed(can.identifier, None, data, can.time)
        if len(data) > 2 and self.use_ext_addr is not False:
            ea = data[0]
            self._try_feed(can.identifier, ea, data[1:], can.time)

    @property
    def count(self):
        # type: () -> int
        """Returns the number of ready ISOTP messages built from the provided
        can frames

        :return: Number of ready ISOTP messages
        """
        return len(self.ready)

    def __len__(self):
        # type: () -> int
        return self.count

    def pop(self, identifier=None, ext_addr=None):
        # type: (Optional[int], Optional[int]) -> Optional[ISOTP]
        """Returns a built ISOTP message

        :param identifier: if not None, only return isotp messages with this
                           destination
        :param ext_addr: if identifier is not None, only return isotp messages
                         with this extended address for destination
        :returns: an ISOTP packet, or None if no message is ready
        """

        if identifier is not None:
            for i in range(len(self.ready)):
                b = self.ready[i]
                iden = b[0]
                ea = b[1]
                if iden == identifier and ext_addr == ea:
                    return ISOTPMessageBuilder._build(self.ready.pop(i),
                                                      self.basecls)
            return None

        if len(self.ready) > 0:
            return ISOTPMessageBuilder._build(self.ready.pop(0), self.basecls)
        return None

    def __iter__(self):
        # type: () -> ISOTPMessageBuilderIter
        return ISOTPMessageBuilderIter(self)

    @staticmethod
    def _build(
            t,  # type: Tuple[int, Optional[int], ISOTPMessageBuilder.Bucket]
            basecls=ISOTP  # type: Type[ISOTP]
    ):
        # type: (...) -> ISOTP
        bucket = t[2]
        data = bucket.ready or b""
        try:
            p = basecls(data)
        except Exception:
            if conf.debug_dissector:
                from scapy.sendrecv import debug
                debug.crashed_on = (basecls, data)
            raise
        if hasattr(p, "rx_id"):
            p.rx_id = t[0]
        if hasattr(p, "rx_ext_address"):
            p.rx_ext_address = t[1]
        if hasattr(p, "tx_id"):
            p.tx_id = bucket.tx_id
        if hasattr(p, "ext_address"):
            p.ext_address = bucket.ext_address
        if hasattr(p, "time"):
            p.time = bucket.time
        return p

    def _feed_first_frame(self, identifier, ea, data, ts):
        # type: (int, Optional[int], bytes, Union[EDecimal, float]) -> bool
        if len(data) < 3:
            # At least 3 bytes are necessary: 2 for length and 1 for data
            return False

        header = struct.unpack('>H', bytes(data[:2]))[0]
        expected_length = header & 0x0fff
        isotp_data = data[2:]
        if expected_length == 0 and len(data) >= 6:
            expected_length = struct.unpack('>I', bytes(data[2:6]))[0]
            isotp_data = data[6:]

        key = (ea, identifier, 1)
        if ea is None:
            self.last_ff = key
        else:
            self.last_ff_ex = key
        self.buckets[key] = self.Bucket(expected_length, isotp_data, ts)
        return True

    def _feed_single_frame(self, identifier, ea, data, ts):
        # type: (int, Optional[int], bytes, Union[EDecimal, float]) -> bool
        if len(data) < 2:
            # At least 2 bytes are necessary: 1 for length and 1 for data
            return False

        length = data[0] & 0x0f
        isotp_data = data[1:length + 1]

        if length > len(isotp_data):
            # CAN frame has less data than expected
            return False

        self.ready.append((identifier, ea,
                           self.Bucket(length, isotp_data, ts)))
        return True

    def _feed_consecutive_frame(self, identifier, ea, data):
        # type: (int, Optional[int], bytes) -> bool
        if len(data) < 2:
            # At least 2 bytes are necessary: 1 for sequence number and
            # 1 for data
            return False

        first_byte = data[0]
        seq_no = first_byte & 0x0f
        isotp_data = data[1:]

        key = (ea, identifier, seq_no)
        bucket = self.buckets.pop(key, None)

        if bucket is None:
            # There is no message constructor waiting for this frame
            return False

        bucket.push(isotp_data)
        if bucket.ready is None:
            # full ISOTP message is not ready yet, put it back in
            # buckets list
            next_seq = (seq_no + 1) % 16
            key = (ea, identifier, next_seq)
            self.buckets[key] = bucket
        else:
            self.ready.append((identifier, ea, bucket))

        return True

    def _feed_flow_control_frame(self, identifier, ea, data):
        # type: (int, Optional[int], bytes) -> bool
        if len(data) < 3:
            # At least 2 bytes are necessary: 1 for sequence number and
            # 1 for data
            return False

        keys = [x for x in (self.last_ff, self.last_ff_ex) if x is not None]
        buckets = [self.buckets.pop(k, None) for k in keys]

        self.last_ff = None
        self.last_ff_ex = None

        if not any(buckets) or not any(keys):
            # There is no message constructor waiting for this frame
            return False

        for key, bucket in zip(keys, buckets):
            if bucket is None:
                continue
            bucket.tx_id = identifier
            bucket.ext_address = ea
            self.buckets[key] = bucket
        return True

    def _try_feed(self, identifier, ea, data, ts):
        # type: (int, Optional[int], bytes, Union[EDecimal, float]) -> None
        first_byte = data[0]
        if len(data) > 1 and first_byte & 0xf0 == N_PCI_SF:
            self._feed_single_frame(identifier, ea, data, ts)
        if len(data) > 2 and first_byte & 0xf0 == N_PCI_FF:
            self._feed_first_frame(identifier, ea, data, ts)
        if len(data) > 1 and first_byte & 0xf0 == N_PCI_CF:
            self._feed_consecutive_frame(identifier, ea, data)
        if len(data) > 1 and first_byte & 0xf0 == N_PCI_FC:
            self._feed_flow_control_frame(identifier, ea, data)


class ISOTPSession(DefaultSession):
    """Defragment ISOTP packets 'on-the-flow'.

    Usage:
        >>> sniff(session=ISOTPSession)
    """

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        self.m = ISOTPMessageBuilder(
            use_ext_address=kwargs.pop("use_ext_address", None),
            rx_id=kwargs.pop("rx_id", None),
            basecls=kwargs.pop("basecls", ISOTP))
        super(ISOTPSession, self).__init__(*args, **kwargs)

    def recv(self, sock: SuperSocket) -> Iterator[Packet]:
        """
        Will be called by sniff() to ask for a packet
        """
        pkt = sock.recv()
        if not pkt:
            return
        self.m.feed(pkt)
        while len(self.m) > 0:
            rcvd = cast(Optional[Packet], self.m.pop())
            if rcvd:
                rcvd = self.process(rcvd)
            if rcvd:
                yield rcvd



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-DCOM]

https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dcom/4a893f3d-bd29-48cd-9f43-d9777a4415b0
"""

import enum
import hashlib
import re
import socket
import uuid

from scapy.config import conf
from scapy.packet import Packet, bind_layers
from scapy.fields import (
    ConditionalField,
    FieldLenField,
    FlagsField,
    LEIntField,
    LELongField,
    LEShortEnumField,
    LEShortField,
    PacketField,
    PacketListField,
    PadField,
    StrLenField,
    StrNullFieldUtf16,
    UUIDField,
    XShortField,
    XStrFixedLenField,
)
from scapy.volatile import RandUUID

from scapy.layers.dcerpc import (
    ComInterface,
    DCE_C_AUTHN_LEVEL,
    DCE_RPC_PROTOCOL_IDENTIFIERS,
    DceRpc5Request,
    find_com_interface,
    find_dcerpc_interface,
    ndr_deserialize1,
    NDRConfFieldListField,
    NDRConfPacketListField,
    NDRConfVarStrNullFieldUtf16,
    NDRFieldListField,
    NDRFullEmbPointerField,
    NDRFullPointerField,
    NDRIntEnumField,
    NDRIntField,
    NDRLongField,
    NDRPacket,
    NDRPacketField,
    NDRSerializeType1PacketField,
    NDRSerializeType1PacketListField,
    NDRShortField,
    NDRSignedIntField,
    RPC_C_AUTHN,
)
from scapy.utils import valid_ip6, valid_ip
from scapy.layers.msrpce.rpcclient import DCERPC_Client, DCERPC_Transport

from scapy.layers.msrpce.raw.ms_dcom import (
    COMVERSION,
    DUALSTRINGARRAY,
    GUID,
    MInterfacePointer,
    ORPCTHAT,
    ORPCTHIS,
    REMINTERFACEREF,
    RemoteCreateInstance_Request,
    RemoteCreateInstance_Response,
    RemoteGetClassObject_Request,
    RemoteGetClassObject_Response,
    RemQueryInterface_Request,
    RemRelease_Request,
    ResolveOxid2_Request,
    ServerAlive2_Request,
    tagCPFLAGS,
)

# Typing
from typing import (
    Any,
    List,
    Dict,
    Optional,
    Tuple,
)


def _uid_to_bytes(x, ndrendian="little"):
    if ndrendian == "little":
        return x.bytes_le
    elif ndrendian == "big":
        return x.bytes
    else:
        raise ValueError("bad ndrendian")


def _uid_from_bytes(x, ndrendian="little"):
    if ndrendian == "little":
        return uuid.UUID(bytes_le=x)
    elif ndrendian == "big":
        return uuid.UUID(bytes=x)
    else:
        raise ValueError("bad ndrendian")


# [MS-DCOM] sect 1.9

CLSID_ActivationContextInfo = uuid.UUID("000001a5-0000-0000-c000-000000000046")
CLSID_ActivationPropertiesIn = uuid.UUID("00000338-0000-0000-c000-000000000046")
CLSID_ActivationPropertiesOut = uuid.UUID("00000339-0000-0000-c000-000000000046")
CLSID_CONTEXT_EXTENSION = uuid.UUID("00000334-0000-0000-c000-000000000046")
CLSID_ContextMarshaler = uuid.UUID("0000033b-0000-0000-c000-000000000046")
CLSID_ERROR_EXTENSION = uuid.UUID("0000031c-0000-0000-c000-000000000046")
CLSID_ErrorObject = uuid.UUID("0000031b-0000-0000-c000-000000000046")
CLSID_InstanceInfo = uuid.UUID("000001ad-0000-0000-c000-000000000046")
CLSID_InstantiationInfo = uuid.UUID("000001ab-0000-0000-c000-000000000046")
CLSID_PropsOutInfo = uuid.UUID("00000339-0000-0000-c000-000000000046")
CLSID_ScmReplyInfo = uuid.UUID("000001b6-0000-0000-c000-000000000046")
CLSID_ScmRequestInfo = uuid.UUID("000001aa-0000-0000-c000-000000000046")
CLSID_SecurityInfo = uuid.UUID("000001a6-0000-0000-c000-000000000046")
CLSID_ServerLocationInfo = uuid.UUID("000001a4-0000-0000-c000-000000000046")
CLSID_SpecialSystemProperties = uuid.UUID("000001b9-0000-0000-c000-000000000046")

# Some special non-interfaces UUIDs:

IID_IActivationPropertiesIn = uuid.UUID("000001A2-0000-0000-C000-000000000046")
IID_IActivationPropertiesOut = uuid.UUID("000001A3-0000-0000-C000-000000000046")
IID_IContext = uuid.UUID("000001c0-0000-0000-C000-000000000046")

# [MS-DCOM] 2.2.22.2.1


class ACTVFLAGS(enum.IntEnum):
    DISABLE_AAA = 0x00000002
    ACTIVATE_32_BIT_SERVER = 0x00000004
    ACTIVATE_64_BIT_SERVER = 0x00000008
    NO_FAILURE_LOG = 0x00000020


class InstantiationInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRPacketField("classId", GUID(), GUID),
        NDRIntField("classCtx", 0),
        NDRIntField("actvflags", 0),
        NDRSignedIntField("fIsSurrogate", 0),
        NDRIntField("cIID", None, size_of="pIID"),
        NDRIntField("instFlag", 0),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "pIID", [GUID()], GUID, count_from=lambda pkt: pkt.cIID
            ),
        ),
        NDRIntField("thisSize", 0),
        NDRPacketField(
            "clientCOMVersion",
            COMVERSION(),
            COMVERSION,
        ),
    ]


# [MS-DCOM] 2.2.22.2.2


class SpecialPropertiesData(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRIntField("dwSessionId", 0xFFFFFFFF),
        NDRSignedIntField("fRemoteThisSessionId", 0),
        NDRSignedIntField("fClientImpersonating", 0),
        NDRSignedIntField("fPartitionIDPresent", 0),
        NDRIntField(
            "dwDefaultAuthnLvl", DCE_C_AUTHN_LEVEL.PKT_INTEGRITY
        ),  # Same than Windows
        NDRPacketField("guidPartition", GUID(), GUID),
        NDRIntField("dwPRTFlags", 0),
        NDRIntField("dwOrigClsctx", 0),
        NDRIntEnumField(
            "dwFlags",
            0,
            {
                0x00000001: "SPD_FLAG_USE_CONSOLE_SESSION",
            },
        ),
        NDRIntField("Reserved1", 0),
        NDRLongField("Reserved2", 0),
        NDRFieldListField(
            "Reserved3", [0, 0, 0, 0, 0], NDRIntField("", 0), count_from=lambda _: 5
        ),
    ]


# [MS-DCOM] 2.2.22.2.3


class InstanceInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("fileName", "")),
        NDRIntField("mode", 0),
        NDRFullEmbPointerField(
            NDRPacketField("ifdROT", MInterfacePointer(), MInterfacePointer),
        ),
        NDRFullEmbPointerField(
            NDRPacketField("ifdStg", MInterfacePointer(), MInterfacePointer),
        ),
    ]


# [MS-DCOM] 2.2.22.2.4


class customREMOTE_REQUEST_SCM_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ClientImpLevel", 2),  # note <33>
        NDRShortField("cRequestedProtseqs", None, size_of="pRequestedProtseqs"),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "pRequestedProtseqs",
                [],
                NDRShortField("", 0),
                size_is=lambda pkt: pkt.cRequestedProtseqs,
            ),
        ),
    ]


class ScmRequestInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRIntField("pdwReserved", 0)),
        NDRFullEmbPointerField(
            NDRPacketField(
                "remoteRequest",
                customREMOTE_REQUEST_SCM_INFO(),
                customREMOTE_REQUEST_SCM_INFO,
            ),
        ),
    ]


# [MS-DCOM] 2.2.22.2.5


class ActivationContextInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRSignedIntField("clientOK", 0),
        NDRSignedIntField("bReserved1", 0),
        NDRIntField("dwReserved1", 0),
        NDRIntField("dwReserved2", 0),
        NDRFullEmbPointerField(
            NDRPacketField("pIFDClientCtx", MInterfacePointer(), MInterfacePointer),
        ),
        NDRFullEmbPointerField(
            NDRPacketField("pIFDPrototypeCtx", MInterfacePointer(), MInterfacePointer),
        ),
    ]


# [MS-DCOM] 2.2.22.2.6


class LocationInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(
            NDRConfVarStrNullFieldUtf16("machineName", None),
        ),
        NDRIntField("processId", 0),
        NDRIntField("apartmentId", 0),
        NDRIntField("contextId", 0),
    ]


# [MS-DCOM] 2.2.22.2.7


class COSERVERINFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwReserved1", 0),
        NDRFullEmbPointerField(NDRConfVarStrNullFieldUtf16("pwszName", "")),
        NDRFullEmbPointerField(NDRIntField("pdwReserved", 0)),
        NDRIntField("dwReserved2", 0),
    ]


class SecurityInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwAuthnFlags", 0),
        NDRFullEmbPointerField(
            NDRPacketField("pServerInfo", COSERVERINFO(), COSERVERINFO),
        ),
        NDRFullPointerField(NDRIntField("pdwReserved", None)),
    ]


class customREMOTE_REPLY_SCM_INFO(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRLongField("Oxid", 0),
        NDRFullEmbPointerField(
            NDRPacketField("pdsaOxidBindings", DUALSTRINGARRAY(), DUALSTRINGARRAY),
        ),
        NDRPacketField("ipidRemUnknown", GUID(), GUID),
        NDRIntField("authnHint", 0),
        NDRPacketField("serverVersion", COMVERSION(), COMVERSION),
    ]


class ScmReplyInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullEmbPointerField(NDRIntField("pdwReserved", 0)),
        NDRFullEmbPointerField(
            NDRPacketField(
                "remoteReply",
                customREMOTE_REPLY_SCM_INFO(),
                customREMOTE_REPLY_SCM_INFO,
            ),
        ),
    ]


# [MS-DCOM] 2.2.22.2.9


class PropsOutInfo(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("cIfs", None, size_of="ppIntfData"),
        NDRFullEmbPointerField(
            NDRConfPacketListField("piid", [], GUID, size_is=lambda pkt: pkt.cIfs)
        ),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "phresults",
                [],
                NDRSignedIntField("phresults", 0),
                size_is=lambda pkt: pkt.cIfs,
            )
        ),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "ppIntfData",
                [],
                MInterfacePointer,
                size_is=lambda pkt: pkt.cIfs,
                ptr_pack=True,
            )
        ),
    ]


# [MS-DCOM] 2.2.22.1


class CustomHeader(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("totalSize", 0),
        NDRIntField("headerSize", 0),
        NDRIntField("dwReserved", 0),
        NDRIntEnumField("destCtx", 2, {2: "MSHCTX_DIFFERENTMACHINE"}),
        NDRIntField("cIfs", None, size_of="pSizes"),
        NDRPacketField("classInfoClsid", GUID(), GUID),
        NDRFullEmbPointerField(
            NDRConfPacketListField(
                "pclsid", [GUID()], GUID, count_from=lambda pkt: pkt.cIfs
            ),
        ),
        NDRFullEmbPointerField(
            NDRConfFieldListField(
                "pSizes", None, NDRIntField("", 0), count_from=lambda pkt: pkt.cIfs
            ),
        ),
        NDRFullEmbPointerField(NDRIntField("pdwReserved", None)),
    ]


class _ActivationPropertiesField(NDRSerializeType1PacketListField):
    def __init__(self, *args, **kwargs):
        kwargs["next_cls_cb"] = self._get_cls_activation
        super(_ActivationPropertiesField, self).__init__(*args, **kwargs)

    def _get_cls_activation(self, pkt, lst, cur, remain):
        # Get all the pcslsid
        pclsid = pkt.CustomHeader[CustomHeader].valueof("pclsid")
        ndrendian = pkt.CustomHeader[CustomHeader].ndrendian
        i = len(lst) + int(bool(cur))
        if i >= len(pclsid):
            return
        # Get the next pclsid we need to process
        next_uid = _uid_from_bytes(bytes(pclsid[i]), ndrendian=ndrendian)
        # [MS-DCOM] 1.9
        cls = {
            CLSID_ActivationContextInfo: ActivationContextInfoData,
            CLSID_InstanceInfo: InstanceInfoData,
            CLSID_InstantiationInfo: InstantiationInfoData,
            CLSID_PropsOutInfo: PropsOutInfo,
            CLSID_ScmReplyInfo: ScmReplyInfoData,
            CLSID_ScmRequestInfo: ScmRequestInfoData,
            CLSID_SecurityInfo: SecurityInfoData,
            CLSID_ServerLocationInfo: LocationInfoData,
            CLSID_SpecialSystemProperties: SpecialPropertiesData,
        }[next_uid]
        return lambda x: ndr_deserialize1(x, cls)


class ActivationPropertiesBlob(Packet):
    fields_desc = [
        FieldLenField(
            "dwSize",
            None,
            fmt="<I",
            length_of="CustomHeader",
            adjust=lambda pkt, x: x
            + pkt.get_field("Property").i2len(pkt, pkt.Property),
        ),
        LEIntField("dwReserved", 0),
        NDRSerializeType1PacketField("CustomHeader", CustomHeader(), CustomHeader),
        _ActivationPropertiesField("Property", []),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.18


class OBJREF(Packet):
    fields_desc = [
        XStrFixedLenField("signature", b"MEOW", length=4),  # :3
        LEIntField("flags", 0x04),
        UUIDField("iid", IID_IActivationPropertiesIn, uuid_fmt=UUIDField.FORMAT_LE),
    ]


# [MS-DCOM] 2.2.18.2


class STDOBJREF(Packet):
    fields_desc = [
        FlagsField(
            "flags",
            0,
            -32,
            {
                0x00001000: "SORF_NOPING",
            },
        ),
        LEIntField("cPublicRefs", 0),
        LELongField("oxid", 0),
        LELongField("oid", 0),
        UUIDField("ipid", None, uuid_fmt=UUIDField.FORMAT_LE),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.18.4


class OBJREF_STANDARD(Packet):
    fields_desc = [
        PacketField("std", STDOBJREF(), STDOBJREF),
        PacketField("saResAddr", DUALSTRINGARRAY(), DUALSTRINGARRAY),
    ]


bind_layers(OBJREF, OBJREF_STANDARD, flags=1)


# [MS-DCOM] 2.2.18.5


class OBJREF_HANDLER(Packet):
    fields_desc = [
        PacketField("std", STDOBJREF(), STDOBJREF),
        UUIDField("clsid", None, uuid_fmt=UUIDField.FORMAT_LE),
        PacketField("saResAddr", DUALSTRINGARRAY(), DUALSTRINGARRAY),
    ]


bind_layers(OBJREF, OBJREF_HANDLER, flags=2)

# [MS-DCOM] 2.2.18.6


class _pObjectDataField(PacketField):
    def m2i(self, pkt, s):
        if pkt.clsid in [CLSID_ActivationPropertiesIn, CLSID_ActivationPropertiesOut]:
            return ActivationPropertiesBlob(s, _parent=pkt)
        elif pkt.clsid == CLSID_ContextMarshaler:
            return Context(s, _parent=pkt)
        return conf.raw_layer(s, _parent=pkt)


class OBJREF_CUSTOM(Packet):
    fields_desc = [
        UUIDField("clsid", CLSID_ActivationPropertiesIn, uuid_fmt=UUIDField.FORMAT_LE),
        LEIntField("cbExtension", 0),
        # The following field is called "reserved" in the spec, but is the size
        # in practice :P
        FieldLenField(
            "reserved",
            None,
            length_of="pObjectData",
            adjust=lambda _, x: x + 8,
            fmt="<I",
        ),
        _pObjectDataField("pObjectData", ActivationPropertiesBlob(), None),
    ]


bind_layers(OBJREF, OBJREF_CUSTOM, flags=4)

# [MS-DCOM] 2.2.18.8


class DATAELEMENT(Packet):
    fields_desc = [
        UUIDField("dataID", None, uuid_fmt=UUIDField.FORMAT_LE),
        FieldLenField("cbSize", None, fmt="<I", length_of="Data"),
        FieldLenField(
            "cbRounded",
            None,
            fmt="<I",
            length_of="Data",
            adjust=lambda _, x: x + (-x % 8),
        ),
        PadField(
            StrLenField("Data", b"", length_from=lambda pkt: pkt.cbSize),
            align=8,
        ),
    ]


# [MS-DCOM] 2.2.18.7


class OBJREF_EXTENDED(Packet):
    fields_desc = [
        PacketField("std", STDOBJREF(), STDOBJREF),
        LEIntField("Signature1", 0x4E535956),
        PacketField("saResAddr", DUALSTRINGARRAY(), DUALSTRINGARRAY),
        LEIntField("nElms", 1),
        LEIntField("Signature2", 0x4E535956),
        # Doc says the array length is always 1
        PacketField("ElmArray", DATAELEMENT(), DATAELEMENT),
    ]


bind_layers(OBJREF, OBJREF_EXTENDED, flags=8)


# [MS-DCOM] 2.2.19.3


class STRINGBINDING(Packet):
    fields_desc = [
        LEShortEnumField("wTowerId", 0, DCE_RPC_PROTOCOL_IDENTIFIERS),
        StrNullFieldUtf16("aNetworkAddr", ""),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.19.4


class SECURITYBINDING(Packet):
    fields_desc = [
        LEShortEnumField("wAuthnSvc", 0, RPC_C_AUTHN),
        ConditionalField(XShortField("Reserved", 0xFFFF), lambda pkt: pkt.wAuthnSvc),
        ConditionalField(
            StrNullFieldUtf16("aPrincName", ""), lambda pkt: pkt.wAuthnSvc
        ),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.20


class PROPMARSHALHEADER(Packet):
    fields_desc = [
        UUIDField("clsid", None, uuid_fmt=UUIDField.FORMAT_LE),
        UUIDField("policyId", None, uuid_fmt=UUIDField.FORMAT_LE),
        FlagsField("flags", 0, -32, tagCPFLAGS),
        FieldLenField("cb", None, length_of="ctxProperty"),
        StrLenField("ctxProperty", b"", length_from=lambda pkt: pkt.cb),
    ]


class Context(Packet):
    fields_desc = [
        LEShortField("MajorVersion", 0x0001),
        LEShortField("MinVersion", 0x0001),
        UUIDField("ContextId", None, uuid_fmt=UUIDField.FORMAT_LE),
        FlagsField(
            "Flags",
            0,
            -32,
            {
                0x00000002: "CTXMSHLFLAGS_BYVAL",
            },
        ),
        LEIntField("Reserved", 0),
        LEIntField("dwNumExtents", 0),
        LEIntField("cbExtents", 0),
        LEIntField("MshlFlags", 0),
        FieldLenField("Count", None, fmt="<I", count_of="PropMarshalHeader"),
        LEIntField("Frozen", 0x00000001),
        PacketListField(
            "PropMarshalHeader", [], PROPMARSHALHEADER, count_from=lambda pkt: pkt.Count
        ),
    ]


# -- DCOM Client and utils --

# Utils


def _ParseStringArray(
    dual: DUALSTRINGARRAY,
) -> Tuple[List[STRINGBINDING], List[SECURITYBINDING]]:
    """
    Process aStringArray in a DUALSTRINGARRAY to extract string bindings and
    security bindings.
    """
    str_fld = PacketListField("", [], STRINGBINDING)
    sec_fld = PacketListField("", [], SECURITYBINDING)
    string = str_fld.getfield(dual, dual.aStringArray[: dual.wSecurityOffset * 2])[1]
    secs = sec_fld.getfield(dual, dual.aStringArray[dual.wSecurityOffset * 2 :])[1]
    if string[-1].wTowerId != 0 or secs[-1].wAuthnSvc != 0:
        raise ValueError("Invalid DUALSTRINGARRAY !")
    return string[:-1], secs[:-1]


def _HashStringBinding(strings: List[STRINGBINDING]):
    """
    Hash a STRINGBINDING list
    """
    return hashlib.sha256(b"".join(bytes(x) for x in strings)).digest()


# Entries.


class IPID_Entry:
    """
    An entry in the IPID table
    [MS-DCOM] 3.1.1.1 Abstract Data Model
    """

    def __init__(self):
        self.ipid: Optional[uuid.UUID] = None
        self.iid: Optional[uuid.UUID] = None
        self.oid: Optional[int] = None
        self.oxid: Optional[int] = None
        self.cPublicRefs: int = 0
        self.cPrivateRefs: int = 0
        self.state: Any = None
        # Additions
        self.iface: Optional[ComInterface] = None


class OID_Entry:
    """
    An entry in the OID table
    [MS-DCOM] 3.1.1.1 Abstract Data Model
    """

    def __init__(self):
        self.oid: Optional[int] = None
        self.oxid: Optional[int] = None
        self.ipids: List[uuid.UUID] = []
        self.hash: Optional[bytes] = None
        self.last_orpc: int = None
        self.garbage_collection: bool = True
        self.state = None


class Resolver_Entry:
    """
    An entry in the Resolver table.
    [MS-DCOM] 3.2.1 Abstract Data Model
    """

    def __init__(self):
        self.hash: Optional[bytes] = None
        self.binds: List[STRINGBINDING] = []
        self.secs: List[SECURITYBINDING] = []
        self.setid: Optional[int] = None
        self.client: Optional[DCERPC_Client] = None


class SETID_Entry:
    """
    An entry in the SETID table.
    [MS-DCOM] 3.2.1 Abstract Data Model
    """

    def __init__(self):
        self.setid: Optional[int] = None
        self.oids: List[int] = []
        self.seq: Optional[int] = None


class OXID_Entry:
    """
    An entry in the OXID table.
    [MS-DCOM] 3.2.1 Abstract Data Model
    """

    def __init__(self):
        self.oxid: Optional[int] = None
        self.bindingInfo: Optional[Tuple[str, int]] = None
        self.authnHint: DCE_C_AUTHN_LEVEL = DCE_C_AUTHN_LEVEL.CONNECT
        self.version: Optional[COMVERSION] = None
        self.ipid_IRemUnknown: Optional[uuid.UUID] = None

    def __repr__(self):
        return f"<OXID_Entry {hex(self.oxid)}>"


class ObjectInstance:
    """
    An reference to an instantiated object.

    This is a helper to manipulate this object and perform calls over it.
    """

    def __init__(self, client: "DCOM_Client", oid: int):
        self.client = client
        self.oid = oid

    def __repr__(self):
        return f"<ObjectInstance {self.oid}>"

    @property
    def valid(self):
        """
        Returns whether the current object still exists
        """
        return self.oid in self.client.OID_table

    @property
    def ndr64(self):
        """
        Whether NDR64 is required to talk to this object
        """
        return self.client.ndr64

    def sr1_req(
        self,
        pkt: NDRPacket,
        iface: ComInterface,
        ssp=None,
        auth_level=None,
        timeout=None,
        **kwargs,
    ):
        """
        Make an ORPC call on this object instance.

        :param iface: the ComInterface to call.
        :param pkt: the request to make.

        :param ssp: (optional) non default SSP to use to connect to the object exporter
        :param auth_level: (optional) non default authn level to use
        :param timeout: (optional) timeout for the connection
        """
        # Look for this object's entry
        try:
            oid_entry = self.client.OID_table[self.oid]
        except KeyError:
            raise ValueError("This object has been released.")

        # Look for the ipid matching the interface required by the user
        ipid = None
        for ipid in oid_entry.ipids:
            ipid_entry = self.client.IPID_table[ipid]
            if ipid_entry.iid == iface.uuid:
                break
        else:
            # Acquire interface on the object
            self.client.AcquireInterface(
                ipid=oid_entry.ipids[0],
                iids=[
                    iface,
                ],
                cPublicRefs=1,
            )

        return self.client.sr1_orpc_req(
            ipid=ipid,
            pkt=pkt,
            ssp=ssp,
            auth_level=auth_level,
            timeout=timeout,
            **kwargs,
        )

    def release(self):
        """
        Call IRemUnknown2::RemRelease to release counts on an object reference.
        """
        for ipid in self.client.OID_table[self.oid].ipids:
            self.client.RemRelease(ipid)


class DCOM_Client(DCERPC_Client):
    """
    A wrapper of DCERPC_Client that adds functions to use COM interfaces.

    :param cid: the client identifier
    """

    IREMUNKNOWN = find_com_interface("IRemUnknown2")

    def __init__(self, cid: GUID = None, verb=True, **kwargs):
        # Pick a random cid to identify this client
        self.cid = cid or GUID(RandUUID().bytes_le)

        # The OXID table kept up-to-date by the client
        self.OXID_table: Dict[int, OXID_Entry] = {}

        # The IPID table kept up-to-date by the client
        self.IPID_table: Dict[int, IPID_Entry] = {}

        # The OID table kept up-to-date by the client
        self.OID_table: Dict[int, OID_Entry] = {}

        # The Resolver table kept up-to-date by the client
        self.Resolver_table: Dict[STRINGBINDING, Resolver_Entry] = {}

        # DCOM defaults to at least PKT_INTEGRITY
        if "auth_level" not in kwargs and "ssp" in kwargs:
            kwargs["auth_level"] = DCE_C_AUTHN_LEVEL.PKT_INTEGRITY

        super(DCOM_Client, self).__init__(
            DCERPC_Transport.NCACN_IP_TCP,
            ndr64=False,
            verb=verb,
            **kwargs,
        )

    def connect(self, host: str, timeout=5):
        """
        Initiate a connection to the object resolver.

        :param host: the host to connect to
        :param timeout: (optional) the connection timeout (default 5)
        """
        # [MS-DCOM] 3.2.4.1.2.1 Determining RPC Binding Information
        binds, _ = ServerAlive2(host)
        host, port = self._ChoseRPCBinding(binds)

        super(DCOM_Client, self).connect(
            host=host,
            port=port,
            timeout=timeout,
        )

    def sr1_req(self, pkt, **kwargs):
        raise NotImplementedError("Cannot use sr1_req on DCOM_Client !")

    def _GetObjectInstance(self, oid: int):
        """
        Internal function to get an ObjectInstance from an oid
        """
        return ObjectInstance(
            client=self,
            oid=oid,
        )

    def _RemoteCreateInstanceOrGetClassObject(
        self,
        clsreq,
        clsresp,
        clsid: uuid.UUID,
        iids: List[ComInterface],
    ) -> ObjectInstance:
        """
        Internal function common to RemoteCreateInstance and RemoteGetClassObject
        """
        if not iids:
            raise ValueError("Must specify at least one interface !")

        # Bind IObjectExporter if not already
        self.bind_or_alter(find_dcerpc_interface("IRemoteSCMActivator"))

        # [MS-DCOM] sect 3.1.2.5.2.3.3 - Issuing the Activation Request

        # Build the activation properties
        ActivationProperties = [
            SpecialPropertiesData(
                # Same as windows
                dwDefaultAuthnLvl=self.auth_level,
                dwOrigClsctx=16,
                dwFlags=2,  # ???
                ndr64=False,
            ),
            InstantiationInfoData(
                classId=GUID(_uid_to_bytes(clsid)),
                classCtx=16,
                actvflags=0,
                fIsSurrogate=0,
                clientCOMVersion=COMVERSION(
                    MajorVersion=5,
                    MinorVersion=7,
                ),
                pIID=[GUID(_uid_to_bytes(x.uuid)) for x in iids],
                ndr64=False,
            ),
            ActivationContextInfoData(
                pIFDClientCtx=MInterfacePointer(
                    abData=OBJREF(iid=IID_IContext)
                    / OBJREF_CUSTOM(
                        clsid=CLSID_ContextMarshaler,
                        pObjectData=Context(
                            ContextId=uuid.UUID("53394e9f-e973-4bf0-a341-154519534fe1"),
                            Flags="CTXMSHLFLAGS_BYVAL",
                        ),
                    ),
                ),
                ndr64=False,
            ),
            SecurityInfoData(
                pServerInfo=COSERVERINFO(
                    pwszName=self.host,
                ),
                ndr64=False,
            ),
            LocationInfoData(ndr64=False),
            ScmRequestInfoData(
                remoteRequest=customREMOTE_REQUEST_SCM_INFO(
                    pRequestedProtseqs=[
                        # Note <51> for Windows Vista and later
                        int(DCERPC_Transport.NCACN_IP_TCP),
                    ]
                ),
                ndr64=False,
            ),
        ]

        # Build CustomHeader
        hdr = CustomHeader(
            pclsid=[
                GUID(_uid_to_bytes(CLSID_SpecialSystemProperties)),
                GUID(_uid_to_bytes(CLSID_InstantiationInfo)),
                GUID(_uid_to_bytes(CLSID_ActivationContextInfo)),
                GUID(_uid_to_bytes(CLSID_SecurityInfo)),
                GUID(_uid_to_bytes(CLSID_ServerLocationInfo)),
                GUID(_uid_to_bytes(CLSID_ScmRequestInfo)),
            ],
            pSizes=[
                # Account for the size of the Type1 header + padding
                len(x) + 16 + (-len(x) % 8)
                for x in ActivationProperties
            ],
            ndr64=False,
        )
        hdr.headerSize = len(hdr) + 16  # 16: size of the Type1 serialization header
        hdr.totalSize = hdr.headerSize + sum(hdr.valueof("pSizes"))

        # Build final request
        pkt = clsreq(
            orpcthis=ORPCTHIS(
                version=COMVERSION(
                    MajorVersion=5,
                    MinorVersion=7,
                ),
                flags=tagCPFLAGS.CPFLAG_PROPAGATE,
                cid=self.cid,
            ),
            pActProperties=MInterfacePointer(
                abData=OBJREF(iid=IID_IActivationPropertiesIn)
                / OBJREF_CUSTOM(
                    clsid=CLSID_ActivationPropertiesIn,
                    pObjectData=ActivationPropertiesBlob(
                        CustomHeader=hdr,
                        Property=ActivationProperties,
                    ),
                ),
            ),
            ndr64=False,
        )

        if isinstance(pkt, RemoteCreateInstance_Request):
            pkt.pUnkOuter = None

        # Send and receive
        resp = super(DCOM_Client, self).sr1_req(pkt)
        if not resp or resp.status != 0:
            raise ValueError("%s failed." % clsreq.__name__)

        entry = OXID_Entry()
        objrefs = []

        # [MS-DCOM] sect 3.2.4.1.1.3 - Updating the Client OXID Table after Activation
        abData = OBJREF(resp.valueof("ppActProperties").abData)
        for prop in abData.pObjectData.Property:
            if ScmReplyInfoData in prop:
                # Information about the object exporter the server found for us
                remoteReply = prop[ScmReplyInfoData].valueof("remoteReply")

                # Get OXID, IPID, COMVERSION, authentication level hint
                entry.oxid = remoteReply.Oxid
                entry.version = remoteReply.serverVersion
                entry.authnHint = DCE_C_AUTHN_LEVEL(remoteReply.authnHint)
                entry.ipid_IRemUnknown = _uid_from_bytes(
                    bytes(remoteReply.ipidRemUnknown), ndrendian=remoteReply.ndrendian
                )

                # Set RPC bindings from the activation request
                binds, _ = _ParseStringArray(remoteReply.valueof("pdsaOxidBindings"))
                entry.bindingInfo = self._ChoseRPCBinding(binds)

            if PropsOutInfo in prop:
                # Information about the interfaces that the client requested
                info = prop[PropsOutInfo]

                # Check that all interfaces were obtained
                phresults = info.valueof("phresults")
                if any(x > 0 for x in phresults):
                    raise ValueError(
                        "Interfaces %s were not obtained !"
                        % [iids[i] for i, x in enumerate(phresults) if x > 0]
                    )

                # Now store the object references for each interface
                for i, ptr in enumerate(info.valueof("ppIntfData")):
                    if phresults[i] == 0:
                        objrefs.append(OBJREF(ptr.abData))
                    else:
                        objrefs.append(None)

        # Update the OXID table
        if entry.oxid not in self.OXID_table:
            self.OXID_table[entry.oxid] = entry

        # Get oid
        oid = objrefs[0].std.oid

        # Add an entry to the IPID table for the RemUnknown
        if entry.ipid_IRemUnknown not in self.IPID_table:
            ipid_entry = IPID_Entry()
            ipid_entry.iface = self.IREMUNKNOWN
            ipid_entry.iid = self.IREMUNKNOWN.uuid
            ipid_entry.oxid = entry.oxid
            ipid_entry.oid = oid
            self.IPID_table[entry.ipid_IRemUnknown] = ipid_entry

        # "For each object reference returned from the activation request for
        # which the corresponding status code indicates success, the client MUST
        # unmarshal the object reference"
        for i, obj in enumerate(objrefs):
            if obj is None:
                continue
            # Unmarshall
            self._UnmarshallObjref(obj, iid=iids[i])

        return self._GetObjectInstance(oid=oid)

    def _UnmarshallObjref(
        self,
        obj: OBJREF,
        iid: Optional[ComInterface] = None,
    ) -> int:
        """
        [MS-DCOM] sect 3.2.4.1.2 - Unmarshaling an Object Reference

        :param iid: "IID specified by the application when unmarshalling the object
            reference" (see [MS-DCOM] sect 4.5)
        """
        # "If the OBJREF_STANDARD flag is set"
        if OBJREF_STANDARD in obj and iid:
            # "the client MUST look up the OXID entry in the OXID
            # table using the OXID from the STDOBJREF"
            try:
                ox = self.OXID_table[obj.std.oxid]
            except KeyError:
                # "If the table entry is not found"

                # "determine the RPC binding information to be used"
                binds, _ = _ParseStringArray(obj.saResAddr)
                host, port = self._ChoseRPCBinding(binds)

                # "issue OXID resolution"
                ox = self.ResolveOxid2(oxid=obj.std.oxid, host=host, port=port)

            # "Next, the client MUST update its tables"
            self._UpdateTables(iid, ox, obj, obj.std)

            # "Finally, the client MUST compare the IID in the OBJREF with the
            # IID specified by the application"
            if obj.iid != iid.uuid:
                # "First, the client SHOULD acquire an object reference of the IID
                # specified by the application"
                self.AcquireInterface(
                    ipid=obj.std.ipid,
                    iids=[
                        iid,
                    ],
                    cPublicRefs=1,
                )

                # "Next, the client MUST release the object reference unmarshaled
                # from the OBJREF"
                self.RemRelease(obj.std.ipid)

            return obj.std.oid
        else:
            obj.show()
            raise NotImplementedError("Non OBJREF_STANDARD ! Please report.")

    def _UpdateTables(
        self,
        iface: ComInterface,
        ox: OXID_Entry,
        obj: OBJREF,
        std: STDOBJREF,
    ) -> None:
        """
        [MS-DCOM] 3.2.4.1.2.3 Updating Client Tables After Unmarshaling
        """
        # [MS-DCOM] 3.2.4.1.2.3.1 Updating the OXID
        if std.oxid not in self.OXID_table:
            self.OXID_table[std.oxid] = ox

        # [MS-DCOM] 3.2.4.1.2.3.2 Updating the OID/IPID/Resolver
        if std.ipid in self.IPID_table:
            self.IPID_table[std.ipid].cPublicRefs += std.cPublicRefs
        else:
            entry = IPID_Entry()
            entry.ipid = std.ipid
            entry.oxid = std.oxid
            entry.oid = std.oid
            entry.iid = obj.iid
            entry.iface = iface
            entry.cPublicRefs = std.cPublicRefs
            if entry.cPublicRefs == 0:
                # "If the STDOBJREF contains a public reference count of zero,
                # the client MUST obtain additional references on the interface"
                raise NotImplementedError("Should acquire additional references !")
            entry.cPrivateRefs = 0
            self.IPID_table[std.ipid] = entry

        if std.oid in self.OID_table:
            oid_entry = self.OID_table[std.oid]
            if std.ipid not in oid_entry.ipids:
                oid_entry.ipids.append(std.ipid)
        else:
            binds, secs = _ParseStringArray(obj.saResAddr)

            oid_entry = OID_Entry()
            oid_entry.oid = std.oid
            oid_entry.oxid = std.oxid
            oid_entry.ipids.append(std.ipid)
            oid_entry.garbage_collection = not std.flags.SORF_NOPING
            oid_entry.hash = _HashStringBinding(binds)
            self.OID_table[std.oid] = oid_entry

            if oid_entry.hash not in self.Resolver_table:
                resolver_entry = Resolver_Entry()
                resolver_entry.setid = 0
                resolver_entry.hash = oid_entry.hash
                resolver_entry.binds = binds
                resolver_entry.secs = secs
                self.Resolver_table[oid_entry.hash] = resolver_entry

    def _ChoseRPCBinding(self, bindings: List[STRINGBINDING]):
        """
        [MS-DCOM] 3.2.4.1.2.1 - Determining RPC Binding Information for OXID Resolution
        """
        # We don't try security bindings, only string ones (connection).
        # We take the first valid one.
        for binding in bindings:
            # Only NCACN_IP_TCP is supported by DCOM
            if binding.wTowerId == DCERPC_Transport.NCACN_IP_TCP:
                # [MS-DCOM] 2.2.19.3
                m = re.match(r"(.*)\[(.*)\]", binding.aNetworkAddr)
                if m:
                    host, port = m.group(1), int(m.group(2))
                else:
                    host, port = binding.aNetworkAddr, 135

                # Check validity of the host/port tuple
                if valid_ip6(host):
                    # IPv6
                    pass
                elif valid_ip(host):
                    # IPv4
                    pass
                else:
                    # Netbios/FQDN
                    try:
                        socket.gethostbyname(host)
                    except Exception:
                        # Resolution failed. Skip.
                        continue

                # Success
                return host, port
        raise ValueError("No valid bindings available !")

    def UnmarshallObjectReference(
        self, mifaceptr: MInterfacePointer, iid: ComInterface
    ):
        """
        [MS-DCOM] 3.2.4.3 Marshaling an Object Reference

        Unmarshall a MInterfacePointer received by the applicative layer.
        """
        oid = self._UnmarshallObjref(obj=OBJREF(mifaceptr.abData), iid=iid)
        return self._GetObjectInstance(oid)

    def ResolveOxid2(
        self, oxid: int, host: Optional[str] = None, port: Optional[int] = None
    ):
        """
        [MS-DCOM] 3.2.4.1.2.2 Issuing the OXID Resolution Request

        :param oxid: the OXID to resolve
        :param host: (optional) connect to a different host
        :param port: (optional) connect to a different port
        """

        if host == self.host and port == self.port:
            host = self.host
            port = self.port
            client = self
        else:
            # Create and connect client
            client = DCOM_Client(
                # Note <85>: Windows uses INTEGRITY
                auth_level=DCE_C_AUTHN_LEVEL.PKT_INTEGRITY,
                ssp=self.ssp,
            )
            client.connect(host, port=port)

        # Bind IObjectExporter if not already
        client.bind_or_alter(find_dcerpc_interface("IObjectExporter"))

        try:
            # Perform ResolveOxid2
            resp = super(DCOM_Client, client).sr1_req(
                ResolveOxid2_Request(
                    pOxid=oxid,
                    arRequestedProtseqs=[
                        DCERPC_Transport.NCACN_IP_TCP,
                    ],
                    ndr64=self.ndr64,
                )
            )
        finally:
            if host != self.host or port != self.port:
                client.close()

        # Entry
        if oxid in self.OXID_table:
            entry = self.OXID_table[oxid]
        else:
            entry = OXID_Entry()

        # Get OXID, IPID, COMVERSION, authentication level hint
        entry.oxid = oxid
        entry.version = resp.pComVersion
        entry.authnHint = DCE_C_AUTHN_LEVEL(resp.pAuthnHint)
        entry.ipid_IRemUnknown = _uid_from_bytes(
            bytes(resp.pipidRemUnknown), ndrendian=resp.ndrendian
        )

        # Set RPC bindings from the oxid request
        binds, _ = _ParseStringArray(resp.valueof("ppdsaOxidBindings"))
        entry.bindingInfo = self._ChoseRPCBinding(binds)

        # Update the OXID table
        if entry.oxid not in self.OXID_table:
            self.OXID_table[entry.oxid] = entry

        return entry

    def RemoteCreateInstance(
        self, clsid: uuid.UUID, iids: List[ComInterface]
    ) -> ObjectInstance:
        """
        Calls IRemoteSCMActivator::RemoteCreateInstance and returns a OXID_Entry
        that points to an instance of the provided class.

        :param clsid: the class ID to initialize
        :param iids: the IDs of the interfaces to request
        """
        return self._RemoteCreateInstanceOrGetClassObject(
            RemoteCreateInstance_Request,
            RemoteCreateInstance_Response,
            clsid,
            iids,
        )

    def RemoteGetClassObject(
        self, clsid: uuid.UUID, iids: List[ComInterface]
    ) -> ObjectInstance:
        """
        Calls IRemoteSCMActivator::RemoteGetClassObject and returns a OXID_Entry
        that points to the factory.

        :param clsid: the class ID to initialize
        :param iids: the IDs of the interfaces to request
        """
        return self._RemoteCreateInstanceOrGetClassObject(
            RemoteGetClassObject_Request,
            RemoteGetClassObject_Response,
            clsid,
            iids,
        )

    def sr1_orpc_req(
        self,
        pkt: NDRPacket,
        ipid: uuid.UUID,
        ssp=None,
        auth_level=None,
        timeout=5,
        **kwargs,
    ):
        """
        Make an ORPC call.

        :param ipid: the reference to a specific interface on an object.
        :param pkt: the request to make.

        :param ssp: (optional) non default SSP to use to connect to the object exporter
        :param auth_level: (optional) non default authn level to use
        :param timeout: (optional) timeout for the connection
        """
        # [MS-DCOM] sect 3.2.4.2

        # 1. look up the object exporter information in the client tables

        try:
            # "The client MUST use the IPID specified by the client application to
            # look up the IPID entry in the IPID table."
            ipid_entry = self.IPID_table[ipid]
        except KeyError:
            raise ValueError("The IPID that was passed is unknown.")

        # "The client MUST then look up the OXID entry"
        oxid_entry = self.OXID_table[ipid_entry.oxid]
        oid_entry = self.OID_table[ipid_entry.oid]
        resolver_entry = self.Resolver_table[oid_entry.hash]

        # Get opnum
        try:
            opnum = pkt.overload_fields[DceRpc5Request]["opnum"]
        except KeyError:
            raise ValueError("This packet is not part of a registered COM interface !")

        # Build ORPC request

        if resolver_entry.client is None:
            # We don't have a client ready, make one.
            resolver_entry.client = DCERPC_Client(
                DCERPC_Transport.NCACN_IP_TCP,
                ssp=ssp or self.ssp,
                auth_level=auth_level or oxid_entry.authnHint,
                verb=self.verb,
            )

            resolver_entry.client.connect(
                host=oxid_entry.bindingInfo[0],
                port=oxid_entry.bindingInfo[1],
                timeout=timeout,
            )

        # Bind the COM interface
        resolver_entry.client.bind_or_alter(ipid_entry.iface)

        # We need to set the NDR very late, after the bind
        pkt.ndr64 = resolver_entry.client.ndr64

        # "The ORPCTHIS and ORPCTHAT structures MUST be marshaled using
        # the NDR [2.0] Transfer Syntax"
        pkt = (
            ORPCTHIS(
                version=oxid_entry.version,
                cid=self.cid,
                ndr64=False,
            )
            / pkt
        )

        # Send/Receive !
        resp = resolver_entry.client.sr1_req(
            pkt,
            opnum=opnum,
            objectuuid=ipid,
            **kwargs,
        )

        return resp[ORPCTHAT].payload

    def AcquireInterface(
        self,
        ipid: uuid.UUID,
        iids: List[ComInterface],
        cPublicRefs: int,
    ):
        """
        [MS-DCOM] 3.2.4.4.3 - Acquiring Additional Interfaces on the Object
        """
        # 1. Look up the OID entry
        ipid_entry = self.IPID_table[ipid]
        oxid_entry = self.OXID_table[ipid_entry.oxid]

        # 2. Perform call
        resp = self.sr1_orpc_req(
            ipid=oxid_entry.ipid_IRemUnknown,
            pkt=RemQueryInterface_Request(
                ripid=GUID(_uid_to_bytes(ipid)),
                cRefs=cPublicRefs,
                cIids=len(iids),
                iids=[GUID(_uid_to_bytes(x.uuid)) for x in iids],
            ),
        )

        # 3. Process answer
        if not resp or resp.status != 0:
            raise ValueError

        # "When the call returns successfully..."
        for i, remqir in enumerate(resp.valueof("ppQIResults")):
            self._UnmarshallObjref(
                OBJREF(iid=iids[i].uuid)
                / OBJREF_STANDARD(std=STDOBJREF(bytes(remqir.std))),
                iid=iids[i],
            )

    def RemRelease(self, ipid: uuid.UUID):
        """
        3.2.4.4.2 Releasing Reference Counts on an Interface
        """

        # 1. Look up the OID entry
        ipid_entry = self.IPID_table[ipid]
        oxid_entry = self.OXID_table[ipid_entry.oxid]
        oid_entry = self.OID_table[ipid_entry.oid]

        # 2. Perform call
        resp = self.sr1_orpc_req(
            ipid=oxid_entry.ipid_IRemUnknown,
            pkt=RemRelease_Request(
                InterfaceRefs=[
                    REMINTERFACEREF(
                        ipid=GUID(_uid_to_bytes(ipid)),
                        cPublicRefs=ipid_entry.cPublicRefs,
                        cPrivateRefs=ipid_entry.cPrivateRefs,
                    )
                ],
            ),
        )

        # 3. Process answer
        if resp and resp.status == 0:
            # "When the call returns successfully..."
            # "It MUST remove the IPID entry from the IPID table."
            del self.IPID_table[ipid]

            # "It MUST remove the IPID from the IPID list in the OID entry."
            oid_entry.ipids.remove(ipid)

            # "If the IPID list of the OID entry is empty, it MUST remove the
            # OID entry from the OID table."
            if not oid_entry.ipids:
                del self.OID_table[ipid_entry.oid]


def ServerAlive2(host, timeout=5) -> Tuple[List[STRINGBINDING], List[SECURITYBINDING]]:
    """
    Call IObjectExporter::ServerAlive2
    """
    client = DCERPC_Client(
        transport=DCERPC_Transport.NCACN_IP_TCP,
        verb=False,
        ndr64=False,
        # "The client MUST NOT specify security on the call"
        auth_level=DCE_C_AUTHN_LEVEL.NONE,
    )
    client.connect(host, port=135, timeout=timeout)

    # Bind IObjectExporter if not already
    client.bind_or_alter(find_dcerpc_interface("IObjectExporter"))

    # Send ServerAlive2 request
    resp = client.sr1_req(ServerAlive2_Request(ndr64=False), timeout=timeout)
    if not resp or resp.status != 0:
        raise ValueError("ServerAlive2 failed !")

    # Parse bindings and security options
    return _ParseStringArray(resp.ppdsaOrBindings.value)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information

"""
Internet Key Exchange Protocol Version 2 (IKEv2), RFC 7296
"""

# scapy.contrib.description = Internet Key Exchange Protocol Version 2 (IKEv2), RFC 7296
# scapy.contrib.status = loads

import struct


# Modified from the original ISAKMP code by Yaron Sheffer <yaronf.ietf@gmail.com>, June 2010.  # noqa: E501

from scapy.packet import (
    Packet,
    Raw,
    bind_bottom_up,
    bind_layers,
    bind_top_down,
    split_bottom_up,
)
from scapy.fields import (
    ByteEnumField,
    ByteField,
    ConditionalField,
    FieldLenField,
    FieldListField,
    FlagsField,
    IP6Field,
    IPField,
    IntField,
    MultiEnumField,
    MultipleTypeField,
    PacketField,
    PacketLenField,
    PacketListField,
    ShortEnumField,
    ShortField,
    StrLenField,
    X3BytesField,
    XByteField,
    XStrFixedLenField,
    XStrLenField,
)
from scapy.layers.x509 import X509_Cert, X509_CRL
from scapy.layers.inet import IP, UDP
from scapy.layers.ipsec import NON_ESP
from scapy.layers.isakmp import ISAKMP
from scapy.sendrecv import sr
from scapy.config import conf
from scapy.volatile import RandString

# see https://www.iana.org/assignments/ikev2-parameters for details
IKEv2AttributeTypes = {
    1: (
        "Encryption",
        {
            1: "DES-IV64",
            2: "DES",
            3: "3DES",
            4: "RC5",
            5: "IDEA",
            6: "CAST",
            7: "Blowfish",
            8: "3IDEA",
            9: "DES-IV32",
            12: "AES-CBC",
            13: "AES-CTR",
            14: "AES-CCM-8",
            15: "AES-CCM-12",
            16: "AES-CCM-16",
            18: "AES-GCM-8ICV",
            19: "AES-GCM-12ICV",
            20: "AES-GCM-16ICV",
            23: "Camellia-CBC",
            24: "Camellia-CTR",
            25: "Camellia-CCM-8ICV",
            26: "Camellia-CCM-12ICV",
            27: "Camellia-CCM-16ICV",
            28: "ChaCha20-Poly1305",
            32: "Kuzneychik-MGM-KTREE",
            33: "MAGMA-MGM-KTREE",
        }
    ),
    2: (
        "PRF",
        {
            1: "PRF_HMAC_MD5",
            2: "PRF_HMAC_SHA1",
            3: "PRF_HMAC_TIGER",
            4: "PRF_AES128_XCBC",
            5: "PRF_HMAC_SHA2_256",
            6: "PRF_HMAC_SHA2_384",
            7: "PRF_HMAC_SHA2_512",
            8: "PRF_AES128_CMAC",
            9: "PRF_HMAC_STREEBOG_512",
        }
    ),
    3: (
        "Integrity",
        {
            1: "HMAC-MD5-96",
            2: "HMAC-SHA1-96",
            3: "DES-MAC",
            4: "KPDK-MD5",
            5: "AES-XCBC-96",
            6: "HMAC-MD5-128",
            7: "HMAC-SHA1-160",
            8: "AES-CMAC-96",
            9: "AES-128-GMAC",
            10: "AES-192-GMAC",
            11: "AES-256-GMAC",
            12: "SHA2-256-128",
            13: "SHA2-384-192",
            14: "SHA2-512-256",
        }
    ),
    4: (
        "GroupDesc",
        {
            1: "768MODPgr",
            2: "1024MODPgr",
            5: "1536MODPgr",
            14: "2048MODPgr",
            15: "3072MODPgr",
            16: "4096MODPgr",
            17: "6144MODPgr",
            18: "8192MODPgr",
            19: "256randECPgr",
            20: "384randECPgr",
            21: "521randECPgr",
            22: "1024MODP160POSgr",
            23: "2048MODP224POSgr",
            24: "2048MODP256POSgr",
            25: "192randECPgr",
            26: "224randECPgr",
            27: "brainpoolP224r1gr",
            28: "brainpoolP256r1gr",
            29: "brainpoolP384r1gr",
            30: "brainpoolP512r1gr",
            31: "curve25519gr",
            32: "curve448gr",
            33: "GOST3410_2012_256",
            34: "GOST3410_2012_512",
        }
    ),
    5: (
        "Extended Sequence Number",
        {
            0: "No ESN",
            1: "ESN"
        }
    ),
}

IKEv2TransformTypes = {
    tf_num: tf_name for tf_name, (tf_num, _) in IKEv2AttributeTypes.items()
}

IKEv2TransformAlgorithms = {
    tf_num: tf_dict for tf_num, (_, tf_dict) in IKEv2AttributeTypes.items()
}

IKEv2ProtocolTypes = {
    1: "IKE",
    2: "AH",
    3: "ESP"
}

IKEv2AuthenticationTypes = {
    0: "Reserved",
    1: "RSA Digital Signature",
    2: "Shared Key Message Integrity Code",
    3: "DSS Digital Signature",
    9: "ECDSA with SHA-256 on the P-256 curve",
    10: "ECDSA with SHA-384 on the P-384 curve",
    11: "ECDSA with SHA-512 on the P-521 curve",
    12: "Generic Secure Password Authentication Method",
    13: "NULL Authentication",
    14: "Digital Signature"
}

IKEv2NotifyMessageTypes = {
    1: "UNSUPPORTED_CRITICAL_PAYLOAD",
    4: "INVALID_IKE_SPI",
    5: "INVALID_MAJOR_VERSION",
    7: "INVALID_SYNTAX",
    9: "INVALID_MESSAGE_ID",
    11: "INVALID_SPI",
    14: "NO_PROPOSAL_CHOSEN",
    17: "INVALID_KE_PAYLOAD",
    24: "AUTHENTICATION_FAILED",
    34: "SINGLE_PAIR_REQUIRED",
    35: "NO_ADDITIONAL_SAS",
    36: "INTERNAL_ADDRESS_FAILURE",
    37: "FAILED_CP_REQUIRED",
    38: "TS_UNACCEPTABLE",
    39: "INVALID_SELECTORS",
    40: "UNACCEPTABLE_ADDRESSES",
    41: "UNEXPECTED_NAT_DETECTED",
    42: "USE_ASSIGNED_HoA",
    43: "TEMPORARY_FAILURE",
    44: "CHILD_SA_NOT_FOUND",
    45: "INVALID_GROUP_ID",
    46: "AUTHORIZATION_FAILED",
    47: "NOTIFY_STATE_NOT_FOUND",
    16384: "INITIAL_CONTACT",
    16385: "SET_WINDOW_SIZE",
    16386: "ADDITIONAL_TS_POSSIBLE",
    16387: "IPCOMP_SUPPORTED",
    16388: "NAT_DETECTION_SOURCE_IP",
    16389: "NAT_DETECTION_DESTINATION_IP",
    16390: "COOKIE",
    16391: "USE_TRANSPORT_MODE",
    16392: "HTTP_CERT_LOOKUP_SUPPORTED",
    16393: "REKEY_SA",
    16394: "ESP_TFC_PADDING_NOT_SUPPORTED",
    16395: "NON_FIRST_FRAGMENTS_ALSO",
    16396: "MOBIKE_SUPPORTED",
    16397: "ADDITIONAL_IP4_ADDRESS",
    16398: "ADDITIONAL_IP6_ADDRESS",
    16399: "NO_ADDITIONAL_ADDRESSES",
    16400: "UPDATE_SA_ADDRESSES",
    16401: "COOKIE2",
    16402: "NO_NATS_ALLOWED",
    16403: "AUTH_LIFETIME",
    16404: "MULTIPLE_AUTH_SUPPORTED",
    16405: "ANOTHER_AUTH_FOLLOWS",
    16406: "REDIRECT_SUPPORTED",
    16407: "REDIRECT",
    16408: "REDIRECTED_FROM",
    16409: "TICKET_LT_OPAQUE",
    16410: "TICKET_REQUEST",
    16411: "TICKET_ACK",
    16412: "TICKET_NACK",
    16413: "TICKET_OPAQUE",
    16414: "LINK_ID",
    16415: "USE_WESP_MODE",
    16416: "ROHC_SUPPORTED",
    16417: "EAP_ONLY_AUTHENTICATION",
    16418: "CHILDLESS_IKEV2_SUPPORTED",
    16419: "QUICK_CRASH_DETECTION",
    16420: "IKEV2_MESSAGE_ID_SYNC_SUPPORTED",
    16421: "IPSEC_REPLAY_COUNTER_SYNC_SUPPORTED",
    16422: "IKEV2_MESSAGE_ID_SYNC",
    16423: "IPSEC_REPLAY_COUNTER_SYNC",
    16424: "SECURE_PASSWORD_METHODS",
    16425: "PSK_PERSIST",
    16426: "PSK_CONFIRM",
    16427: "ERX_SUPPORTED",
    16428: "IFOM_CAPABILITY",
    16429: "SENDER_REQUEST_ID",
    16430: "IKEV2_FRAGMENTATION_SUPPORTED",
    16431: "SIGNATURE_HASH_ALGORITHMS",
    16432: "CLONE_IKE_SA_SUPPORTED",
    16433: "CLONE_IKE_SA",
    16434: "IV2_NOTIFY_PUZZLE",
    16435: "IV2_NOTIFY_USE_PPK",
    16436: "IV2_NOTIFY_PPK_IDENTITY",
    16437: "IV2_NOTIFY_NO_PPK_AUTH",
    16438: "IV2_NOTIFY_INTERMEDIATE_EXCHANGE_SUPPORTED",
    16439: "IV2_NOTIFY_IP4_ALLOWED",
    16440: "IV2_NOTIFY_IP6_ALLOWED",
    16441: "IV2_NOTIFY_ADDITIONAL_KEY_EXCHANGE",
    16442: "IV2_NOTIFY_USE_AGGFRAG",
}

IKEv2GatewayIDTypes = {
    1: "IPv4_addr",
    2: "IPv6_addr",
    3: "FQDN"
}

IKEv2CertificateEncodings = {
    1: "PKCS #7 wrapped X.509 certificate",
    2: "PGP Certificate",
    3: "DNS Signed Key",
    4: "X.509 Certificate - Signature",
    6: "Kerberos Token",
    7: "Certificate Revocation List (CRL)",
    8: "Authority Revocation List (ARL)",
    9: "SPKI Certificate",
    10: "X.509 Certificate - Attribute",
    11: "Raw RSA Key",
    12: "Hash and URL of X.509 certificate",
    13: "Hash and URL of X.509 bundle"
}

IKEv2TrafficSelectorTypes = {
    7: "TS_IPV4_ADDR_RANGE",
    8: "TS_IPV6_ADDR_RANGE",
    9: "TS_FC_ADDR_RANGE"
}

IKEv2ConfigurationPayloadCFGTypes = {
    1: "CFG_REQUEST",
    2: "CFG_REPLY",
    3: "CFG_SET",
    4: "CFG_ACK"
}

IKEv2ConfigurationAttributeTypes = {
    1: "INTERNAL_IP4_ADDRESS",
    2: "INTERNAL_IP4_NETMASK",
    3: "INTERNAL_IP4_DNS",
    4: "INTERNAL_IP4_NBNS",
    6: "INTERNAL_IP4_DHCP",
    7: "APPLICATION_VERSION",
    8: "INTERNAL_IP6_ADDRESS",
    10: "INTERNAL_IP6_DNS",
    12: "INTERNAL_IP6_DHCP",
    13: "INTERNAL_IP4_SUBNET",
    14: "SUPPORTED_ATTRIBUTES",
    15: "INTERNAL_IP6_SUBNET",
    16: "MIP6_HOME_PREFIX",
    17: "INTERNAL_IP6_LINK",
    18: "INTERNAL_IP6_PREFIX",
    19: "HOME_AGENT_ADDRESS",
    20: "P_CSCF_IP4_ADDRESS",
    21: "P_CSCF_IP6_ADDRESS",
    22: "FTT_KAT",
    23: "EXTERNAL_SOURCE_IP4_NAT_INFO",
    24: "TIMEOUT_PERIOD_FOR_LIVENESS_CHECK",
    25: "INTERNAL_DNS_DOMAIN",
    26: "INTERNAL_DNSSEC_TA"
}

IPProtocolIDs = {
    0: "All protocols",
    1: "Internet Control Message Protocol",
    2: "Internet Group Management Protocol",
    3: "Gateway-to-Gateway Protocol",
    4: "IP in IP (encapsulation)",
    5: "Internet Stream Protocol",
    6: "Transmission Control Protocol",
    7: "Core-based trees",
    8: "Exterior Gateway Protocol",
    9: "Interior Gateway Protocol (any private interior gateway (used by Cisco for their IGRP))",  # noqa: E501
    10: "BBN RCC Monitoring",
    11: "Network Voice Protocol",
    12: "Xerox PUP",
    13: "ARGUS",
    14: "EMCON",
    15: "Cross Net Debugger",
    16: "Chaos",
    17: "User Datagram Protocol",
    18: "Multiplexing",
    19: "DCN Measurement Subsystems",
    20: "Host Monitoring Protocol",
    21: "Packet Radio Measurement",
    22: "XEROX NS IDP",
    23: "Trunk-1",
    24: "Trunk-2",
    25: "Leaf-1",
    26: "Leaf-2",
    27: "Reliable Datagram Protocol",
    28: "Internet Reliable Transaction Protocol",
    29: "ISO Transport Protocol Class 4",
    30: "Bulk Data Transfer Protocol",
    31: "MFE Network Services Protocol",
    32: "MERIT Internodal Protocol",
    33: "Datagram Congestion Control Protocol",
    34: "Third Party Connect Protocol",
    35: "Inter-Domain Policy Routing Protocol",
    36: "Xpress Transport Protocol",
    37: "Datagram Delivery Protocol",
    38: "IDPR Control Message Transport Protocol",
    39: "TP++ Transport Protocol",
    40: "IL Transport Protocol",
    41: "IPv6 Encapsulation",
    42: "Source Demand Routing Protocol",
    43: "Routing Header for IPv6",
    44: "Fragment Header for IPv6",
    45: "Inter-Domain Routing Protocol",
    46: "Resource Reservation Protocol",
    47: "Generic Routing Encapsulation",
    48: "Mobile Host Routing Protocol",
    49: "BNA",
    50: "Encapsulating Security Payload",
    51: "Authentication Header",
    52: "Integrated Net Layer Security Protocol",
    53: "SwIPe",
    54: "NBMA Address Resolution Protocol",
    55: "IP Mobility (Min Encap)",
    56: "Transport Layer Security Protocol (using Kryptonet key management)",
    57: "Simple Key-Management for Internet Protocol",
    58: "ICMP for IPv6",
    59: "No Next Header for IPv6",
    60: "Destination Options for IPv6",
    61: "Any host internal protocol",
    62: "CFTP",
    63: "Any local network",
    64: "SATNET and Backroom EXPAK",
    65: "Kryptolan",
    66: "MIT Remote Virtual Disk Protocol",
    67: "Internet Pluribus Packet Core",
    68: "Any distributed file system",
    69: "SATNET Monitoring",
    70: "VISA Protocol",
    71: "Internet Packet Core Utility",
    72: "Computer Protocol Network Executive",
    73: "Computer Protocol Heart Beat",
    74: "Wang Span Network",
    75: "Packet Video Protocol",
    76: "Backroom SATNET Monitoring",
    77: "SUN ND PROTOCOL-Temporary",
    78: "WIDEBAND Monitoring",
    79: "WIDEBAND EXPAK",
    80: "International Organization for Standardization Internet Protocol",
    81: "Versatile Message Transaction Protocol",
    82: "Secure Versatile Message Transaction Protocol",
    83: "VINES",
    84: "Internet Protocol Traffic Manager",
    85: "NSFNET-IGP",
    86: "Dissimilar Gateway Protocol",
    87: "TCF",
    88: "EIGRP",
    89: "Open Shortest Path First",
    90: "Sprite RPC Protocol",
    91: "Locus Address Resolution Protocol",
    92: "Multicast Transport Protocol",
    93: "AX.25",
    94: "IP-within-IP Encapsulation Protocol",
    95: "Mobile Internetworking Control Protocol",
    96: "Semaphore Communications Sec. Pro",
    97: "Ethernet-within-IP Encapsulation",
    98: "Encapsulation Header",
    99: "Any private encryption scheme",
    100: "GMTP",
    101: "Ipsilon Flow Management Protocol",
    102: "PNNI over IP",
    103: "Protocol Independent Multicast",
    104: "IBM's ARIS (Aggregate Route IP Switching) Protocol",
    105: "SCPS (Space Communications Protocol Standards)",
    106: "QNX",
    107: "Active Networks",
    108: "IP Payload Compression Protocol",
    109: "Sitara Networks Protocol",
    110: "Compaq Peer Protocol",
    111: "IPX in IP",
    112: "Virtual Router Redundancy Protocol, Common Address Redundancy Protocol (not IANA assigned)",  # noqa: E501
    113: "PGM Reliable Transport Protocol",
    114: "Any 0-hop protocol",
    115: "Layer Two Tunneling Protocol Version 3",
    116: "D-II Data Exchange (DDX)",
    117: "Interactive Agent Transfer Protocol",
    118: "Schedule Transfer Protocol",
    119: "SpectraLink Radio Protocol",
    120: "Universal Transport Interface Protocol",
    121: "Simple Message Protocol",
    122: "Simple Multicast Protocol",
    123: "Performance Transparency Protocol",
    124: "Intermediate System to Intermediate System (IS-IS) Protocol over IPv4",  # noqa: E501
    125: "Flexible Intra-AS Routing Environment",
    126: "Combat Radio Transport Protocol",
    127: "Combat Radio User Datagram",
    128: "Service-Specific Connection-Oriented Protocol in a Multilink and Connectionless Environment",  # noqa: E501
    129: "IPLT",
    130: "Secure Packet Shield",
    131: "Private IP Encapsulation within IP",
    132: "Stream Control Transmission Protocol",
    133: "Fibre Channel",
    134: "Reservation Protocol (RSVP) End-to-End Ignore",
    135: "Mobility Extension Header for IPv6",
    136: "Lightweight User Datagram Protocol",
    137: "Multiprotocol Label Switching Encapsulated in IP",
    138: "MANET Protocols",
    139: "Host Identity Protocol",
    140: "Site Multihoming by IPv6 Intermediation",
    141: "Wrapped Encapsulating Security Payload",
    142: "Robust Header Compression",
}

IKEv2PayloadTypes = {
    0: "None",
    2: "Proposal",   # used only inside the SA payload
    3: "Transform",  # used only inside the SA payload
    33: "SA",
    34: "KE",
    35: "IDi",
    36: "IDr",
    37: "CERT",
    38: "CERTREQ",
    39: "AUTH",
    40: "Nonce",
    41: "Notify",
    42: "Delete",
    43: "VendorID",
    44: "TSi",
    45: "TSr",
    46: "Encrypted",
    47: "CP",
    48: "EAP",
    49: "GSPM",
    50: "IDg",
    51: "GSA",
    52: "KD",
    53: "Encrypted_Fragment",
    54: "PS"
}


IKEv2ExchangeTypes = {
    34: "IKE_SA_INIT",
    35: "IKE_AUTH",
    36: "CREATE_CHILD_SA",
    37: "INFORMATIONAL",
    38: "IKE_SESSION_RESUME",
    43: "IKE_INTERMEDIATE"
}


class _IKEv2_Packet(Packet):
    def default_payload_class(self, payload):
        return IKEv2_Payload if self.next_payload else conf.raw_layer


class IKEv2(_IKEv2_Packet):  # rfc4306
    name = "IKEv2"
    fields_desc = [
        XStrFixedLenField("init_SPI", "", 8),
        XStrFixedLenField("resp_SPI", "", 8),
        ByteEnumField("next_payload", 0, IKEv2PayloadTypes),
        XByteField("version", 0x20),
        ByteEnumField("exch_type", 0, IKEv2ExchangeTypes),
        FlagsField("flags", 0, 8, ["res0", "res1", "res2", "Initiator", "Version", "Response", "res6", "res7"]),  # noqa: E501
        IntField("id", 0),
        IntField("length", None)  # Length of total message: packets + all payloads  # noqa: E501
    ]

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 18:
            version = struct.unpack("!B", _pkt[17:18])[0]
            if version < 0x20:
                return ISAKMP
        return cls

    def answers(self, other):
        if isinstance(other, IKEv2):
            if other.init_SPI == self.init_SPI:
                return 1
        return 0

    def post_build(self, p, pay):
        p += pay
        if self.length is None:
            p = p[:24] + struct.pack("!I", len(p)) + p[28:]
        return p


class IKEv2_Key_Length_Attribute(IntField):
    # We only support the fixed-length Key Length attribute (the only one currently defined)  # noqa: E501
    def __init__(self, name):
        IntField.__init__(self, name, 0x800E0000)

    def i2h(self, pkt, x):
        return IntField.i2h(self, pkt, x & 0xFFFF)

    def h2i(self, pkt, x):
        return IntField.h2i(self, pkt, (x if x is not None else 0) | 0x800E0000)  # noqa: E501


class IKEv2_Payload(_IKEv2_Packet):
    name = "IKEv2 Payload"
    fields_desc = [
        ByteEnumField("next_payload", None, IKEv2PayloadTypes),
        FlagsField("flags", 0, 8, {0x80: "critical"}),
        ShortField("length", None),
        XStrLenField("load", "", length_from=lambda pkt: pkt.length - 4),
    ]

    def post_build(self, pkt, pay):
        if self.length is None:
            pkt = pkt[:2] + struct.pack("!H", len(pkt)) + pkt[4:]
        return pkt + pay


class IKEv2_Transform(IKEv2_Payload):
    name = "IKEv2 Transform"
    fields_desc = IKEv2_Payload.fields_desc[:2] + [
        ShortField("length", 8),  # can't be None, because 'key_length' depends on it
        ByteEnumField("transform_type", None, IKEv2TransformTypes),
        ByteField("res2", 0),
        MultiEnumField("transform_id", None, IKEv2TransformAlgorithms, depends_on=lambda pkt: pkt.transform_type, fmt="H"),  # noqa: E501
        ConditionalField(IKEv2_Key_Length_Attribute("key_length"), lambda pkt: pkt.length > 8),  # noqa: E501
    ]


class IKEv2_Proposal(IKEv2_Payload):
    name = "IKEv2 Proposal"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteField("proposal", 1),
        ByteEnumField("proto", 1, IKEv2ProtocolTypes),
        FieldLenField("SPIsize", None, "SPI", "B"),
        ByteField("trans_nb", None),
        XStrLenField("SPI", "", length_from=lambda pkt: pkt.SPIsize),
        PacketLenField("trans", conf.raw_layer(), IKEv2_Transform, length_from=lambda pkt: pkt.length - 8 - pkt.SPIsize),  # noqa: E501
    ]


class IKEv2_AUTH(IKEv2_Payload):
    name = "IKEv2 Authentication"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("auth_type", None, IKEv2AuthenticationTypes),
        X3BytesField("res2", 0),
        XStrLenField("load", "", length_from=lambda pkt: pkt.length - 8),
    ]


class IKEv2_VendorID(IKEv2_Payload):
    name = "IKEv2 Vendor ID"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        XStrLenField("vendorID", "", length_from=lambda pkt: pkt.length - 4),
    ]


class TrafficSelector(Packet):
    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 16:
            ts_type = struct.unpack("!B", _pkt[0:1])[0]
            if ts_type == 7:
                return IPv4TrafficSelector
            elif ts_type == 8:
                return IPv6TrafficSelector
            elif ts_type == 9:
                return EncryptedTrafficSelector
            else:
                return RawTrafficSelector
        return IPv4TrafficSelector

    def extract_padding(self, s):
        return '', s


class IPv4TrafficSelector(TrafficSelector):
    name = "IKEv2 IPv4 Traffic Selector"
    fields_desc = [
        ByteEnumField("TS_type", 7, IKEv2TrafficSelectorTypes),
        ByteEnumField("IP_protocol_ID", None, IPProtocolIDs),
        ShortField("length", 16),
        ShortField("start_port", 0),
        ShortField("end_port", 65535),
        IPField("starting_address_v4", "192.168.0.1"),
        IPField("ending_address_v4", "192.168.0.255"),
    ]


class IPv6TrafficSelector(TrafficSelector):
    name = "IKEv2 IPv6 Traffic Selector"
    fields_desc = [
        ByteEnumField("TS_type", 8, IKEv2TrafficSelectorTypes),
        ByteEnumField("IP_protocol_ID", None, IPProtocolIDs),
        ShortField("length", 20),
        ShortField("start_port", 0),
        ShortField("end_port", 65535),
        IP6Field("starting_address_v6", "2001::"),
        IP6Field("ending_address_v6", "2001::"),
    ]


class EncryptedTrafficSelector(TrafficSelector):
    name = "IKEv2 Encrypted Traffic Selector"
    fields_desc = [
        ByteEnumField("TS_type", 9, IKEv2TrafficSelectorTypes),
        ByteEnumField("IP_protocol_ID", None, IPProtocolIDs),
        ShortField("length", 16),
        ByteField("res", 0),
        X3BytesField("starting_address_FC", 0),
        ByteField("res2", 0),
        X3BytesField("ending_address_FC", 0),
        ByteField("starting_R_CTL", 0),
        ByteField("ending_R_CTL", 0),
        ByteField("starting_type", 0),
        ByteField("ending_type", 0),
    ]


class RawTrafficSelector(TrafficSelector):
    name = "IKEv2 Encrypted Traffic Selector"
    fields_desc = [
        ByteEnumField("TS_type", None, IKEv2TrafficSelectorTypes),
        ByteEnumField("IP_protocol_ID", None, IPProtocolIDs),
        FieldLenField("length", None, "load", "H", adjust=lambda pkt, x: x + 4),
        PacketField("load", "", Raw)
    ]


class IKEv2_TSi(IKEv2_Payload):
    name = "IKEv2 Traffic Selector - Initiator"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        FieldLenField("number_of_TSs", None, fmt="B",
                      count_of="traffic_selector"),
        X3BytesField("res2", 0),
        PacketListField("traffic_selector", None, TrafficSelector,
                        length_from=lambda pkt: pkt.length - 8,
                        count_from=lambda pkt: pkt.number_of_TSs),
    ]


class IKEv2_TSr(IKEv2_Payload):
    name = "IKEv2 Traffic Selector - Responder"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        FieldLenField("number_of_TSs", None, fmt="B",
                      count_of="traffic_selector"),
        X3BytesField("res2", 0),
        PacketListField("traffic_selector", None, TrafficSelector,
                        length_from=lambda pkt: pkt.length - 8,
                        count_from=lambda pkt: pkt.number_of_TSs),
    ]


class IKEv2_Delete(IKEv2_Payload):
    name = "IKEv2 Delete"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("proto", None, {0: "Reserved", 1: "IKE", 2: "AH", 3: "ESP"}),  # noqa: E501
        FieldLenField("SPIsize", None, "SPI", "B"),
        ShortField("SPInum", 0),
        FieldListField("SPI", [],
                       XStrLenField("", "", length_from=lambda pkt: pkt.SPIsize),
                       count_from=lambda pkt: pkt.SPInum)
    ]


class IKEv2_SA(IKEv2_Payload):
    name = "IKEv2 SA"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        PacketLenField("prop", conf.raw_layer(), IKEv2_Proposal, length_from=lambda pkt: pkt.length - 4),  # noqa: E501
    ]


class IKEv2_Nonce(IKEv2_Payload):
    name = "IKEv2 Nonce"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        XStrLenField("nonce", "", length_from=lambda pkt: pkt.length - 4),
    ]


class IKEv2_Notify(IKEv2_Payload):
    name = "IKEv2 Notify"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("proto", None, IKEv2ProtocolTypes),
        FieldLenField("SPIsize", None, "SPI", "B"),
        ShortEnumField("type", 0, IKEv2NotifyMessageTypes),
        XStrLenField("SPI", "", length_from=lambda pkt: pkt.SPIsize),
        ConditionalField(
            XStrLenField("notify", "",
                         length_from=lambda pkt: pkt.length - 8 - pkt.SPIsize),
            lambda pkt: pkt.type not in (16407, 16408)
        ),
        ConditionalField(
            # REDIRECT, REDIRECTED_FROM  (RFC 5685)
            ByteEnumField("gw_id_type", 1, IKEv2GatewayIDTypes),
            lambda pkt: pkt.type in (16407, 16408)
        ),
        ConditionalField(
            # REDIRECT, REDIRECTED_FROM  (RFC 5685)
            FieldLenField("gw_id_len", None, "gw_id", "B"),
            lambda pkt: pkt.type in (16407, 16408)
        ),
        ConditionalField(
            # REDIRECT, REDIRECTED_FROM  (RFC 5685)
            MultipleTypeField(
                [
                    (IPField("gw_id", "127.0.0.1"), lambda x: x.gw_id_type == 1),
                    (IP6Field("gw_id", "::1"), lambda x: x.gw_id_type == 2),
                ],
                StrLenField("gw_id", "", length_from=lambda x: x.gw_id_len)
            ),
            lambda pkt: pkt.type in (16407, 16408)
        ),
        ConditionalField(
            # REDIRECT  (RFC 5685)
            XStrLenField("nonce", "", length_from=lambda x:x.length - 10 - x.gw_id_len),
            lambda pkt: pkt.type == 16407
        )
    ]


class IKEv2_KE(IKEv2_Payload):
    name = "IKEv2 Key Exchange"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ShortEnumField("group", 0, IKEv2TransformAlgorithms[4]),
        ShortField("res2", 0),
        XStrLenField("ke", "", length_from=lambda pkt: pkt.length - 8),
    ]


class IKEv2_IDi(IKEv2_Payload):  # RFC 7296, section 3.5
    name = "IKEv2 Identification - Initiator"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("IDtype", 1, {1: "IPv4_addr", 2: "FQDN", 3: "Email_addr", 5: "IPv6_addr", 11: "Key"}),  # noqa: E501
        X3BytesField("res2", 0),
        MultipleTypeField(
            [
                (IPField("ID", "127.0.0.1"), lambda pkt: pkt.IDtype == 1),
                (IP6Field("ID", "::1"), lambda pkt: pkt.IDtype == 5),
            ],
            XStrLenField("ID", "", length_from=lambda pkt: pkt.length - 8),
        )
    ]


class IKEv2_IDr(IKEv2_Payload):  # RFC 7296, section 3.5
    name = "IKEv2 Identification - Responder"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("IDtype", 1, {1: "IPv4_addr", 2: "FQDN", 3: "Email_addr", 5: "IPv6_addr", 11: "Key"}),  # noqa: E501
        X3BytesField("res2", 0),
        MultipleTypeField(
            [
                (IPField("ID", "127.0.0.1"), lambda pkt: pkt.IDtype == 1),
                (IP6Field("ID", "::1"), lambda pkt: pkt.IDtype == 5),
            ],
            XStrLenField("ID", "", length_from=lambda pkt: pkt.length - 8),
        )
    ]


class IKEv2_Encrypted(IKEv2_Payload):
    name = "IKEv2 Encrypted and Authenticated"


class ConfigurationAttribute(Packet):
    name = "IKEv2 Configuration Attribute"
    fields_desc = [
        ShortEnumField("type", 1, IKEv2ConfigurationAttributeTypes),
        FieldLenField("length", None, "value", "H"),
        MultipleTypeField(
            [
                (IPField("value", "127.0.0.1"),
                 lambda pkt: pkt.length == 4 and pkt.type in (1, 2, 3, 4, 6, 20)),
                (IP6Field("value", "::1"),
                 lambda pkt: pkt.length == 16 and pkt.type in (10, 12, 21)),
            ],
            XStrLenField("value", "", length_from=lambda pkt: pkt.length),
        )
    ]

    def extract_padding(self, s):
        return b'', s


class IKEv2_CP(IKEv2_Payload):  # RFC 7296, section 3.15
    name = "IKEv2 Configuration"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("CFGType", 1, IKEv2ConfigurationPayloadCFGTypes),
        X3BytesField("res2", 0),
        PacketListField("attributes", None, ConfigurationAttribute,
                        length_from=lambda pkt: pkt.length - 8),
    ]


class IKEv2_Encrypted_Fragment(IKEv2_Payload):
    name = "IKEv2 Encrypted and Authenticated Fragment"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ShortField("frag_number", 1),
        ShortField("frag_total", 1),
        XStrLenField("load", "", length_from=lambda pkt: pkt.length - 8),
    ]


class IKEv2_CERTREQ(IKEv2_Payload):
    name = "IKEv2 Certificate Request"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("cert_encoding", 0, IKEv2CertificateEncodings),
        XStrLenField("cert_authority", "", length_from=lambda pkt: pkt.length - 5),
    ]


class IKEv2_CERT(IKEv2_Payload):
    name = "IKEv2 Certificate"
    fields_desc = IKEv2_Payload.fields_desc[:3] + [
        ByteEnumField("cert_encoding", 4, IKEv2CertificateEncodings),
        MultipleTypeField(
            [
                (PacketLenField("cert_data", X509_Cert(), X509_Cert,
                                length_from=lambda pkt: pkt.length - 5),
                 lambda pkt: pkt.cert_encoding == 4),
                (PacketLenField("cert_data", X509_CRL(), X509_CRL,
                                length_from=lambda pkt: pkt.length - 5),
                 lambda pkt: pkt.cert_encoding == 7)
            ],
            XStrLenField("cert_data", "", length_from=lambda pkt: pkt.length - 5),
        )
    ]


# TODO: the following payloads are not fully dissected yet

class IKEv2_EAP(IKEv2_Payload):
    name = "IKEv2 Extensible Authentication"


class IKEv2_GSPM(IKEv2_Payload):
    name = "Generic Secure Password Method"


class IKEv2_IDg(IKEv2_Payload):
    name = "Group Identification"


class IKEv2_GSA(IKEv2_Payload):
    name = "Group Security Association"


class IKEv2_KD(IKEv2_Payload):
    name = "Key Download"


class IKEv2_PS(IKEv2_Payload):
    name = "Puzzle Solution"


# bind all IKEv2 payload classes together
bind_layers(_IKEv2_Packet, IKEv2_Proposal, next_payload=2)
bind_layers(_IKEv2_Packet, IKEv2_Transform, next_payload=3)
bind_layers(_IKEv2_Packet, IKEv2_SA, next_payload=33)
bind_layers(_IKEv2_Packet, IKEv2_KE, next_payload=34)
bind_layers(_IKEv2_Packet, IKEv2_IDi, next_payload=35)
bind_layers(_IKEv2_Packet, IKEv2_IDr, next_payload=36)
bind_layers(_IKEv2_Packet, IKEv2_CERT, next_payload=37)
bind_layers(_IKEv2_Packet, IKEv2_CERTREQ, next_payload=38)
bind_layers(_IKEv2_Packet, IKEv2_AUTH, next_payload=39)
bind_layers(_IKEv2_Packet, IKEv2_Nonce, next_payload=40)
bind_layers(_IKEv2_Packet, IKEv2_Notify, next_payload=41)
bind_layers(_IKEv2_Packet, IKEv2_Delete, next_payload=42)
bind_layers(_IKEv2_Packet, IKEv2_VendorID, next_payload=43)
bind_layers(_IKEv2_Packet, IKEv2_TSi, next_payload=44)
bind_layers(_IKEv2_Packet, IKEv2_TSr, next_payload=45)
bind_layers(_IKEv2_Packet, IKEv2_Encrypted, next_payload=46)
bind_layers(_IKEv2_Packet, IKEv2_CP, next_payload=47)
bind_layers(_IKEv2_Packet, IKEv2_EAP, next_payload=48)
bind_layers(_IKEv2_Packet, IKEv2_GSPM, next_payload=49)
bind_layers(_IKEv2_Packet, IKEv2_IDg, next_payload=50)
bind_layers(_IKEv2_Packet, IKEv2_GSA, next_payload=51)
bind_layers(_IKEv2_Packet, IKEv2_KD, next_payload=52)
bind_layers(_IKEv2_Packet, IKEv2_Encrypted_Fragment, next_payload=53)
bind_layers(_IKEv2_Packet, IKEv2_PS, next_payload=54)

# the upper bindings for port 500 to ISAKMP are handled by IKEv2.dispatch_hook
split_bottom_up(UDP, ISAKMP, dport=500)
split_bottom_up(UDP, ISAKMP, sport=500)

bind_bottom_up(UDP, IKEv2, dport=500)
bind_bottom_up(UDP, IKEv2, sport=500)
bind_top_down(UDP, IKEv2, dport=500, sport=500)

split_bottom_up(NON_ESP, ISAKMP)
bind_bottom_up(NON_ESP, IKEv2)


def ikev2scan(ip, **kwargs):
    """Send a IKEv2 SA to an IP and wait for answers."""
    return sr(IP(dst=ip) / UDP() / IKEv2(init_SPI=RandString(8),
                                         exch_type=34) / IKEv2_SA(prop=IKEv2_Proposal()), **kwargs)  # noqa: E501



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Nils Weiss <nils@we155.de>
# Copyright (C) Alexander Schroeder <alexander1.schroeder@st.othr.de>


import getopt
import sys
import signal
import re
import threading

from ast import literal_eval

from scapy.config import conf
from scapy.consts import LINUX

# Typing imports
from typing import (
    Tuple,
    Optional,
    Any,
)

if not LINUX or conf.use_pypy:
    conf.contribs['CANSocket'] = {'use-python-can': True}

from scapy.contrib.cansocket import CANSocket, PYTHON_CAN   # noqa: E402
from scapy.contrib.isotp import isotp_scan  # noqa: E402


def usage(is_error):
    # type: (bool) -> None
    print('''usage:\tisotpscanner [-i interface] [-c channel]
                [-a python-can_args] [-n NOISE_LISTEN_TIME] [-t SNIFF_TIME]
                [-x|--extended] [-C|--piso] [-v|--verbose] [-h|--help]
                [-s start] [-e end]\n
    Scan for open ISOTP-Sockets.\n
    required arguments:
    -c, --channel         python-can channel or Linux SocketCAN interface name
    -s, --start           Start scan at this identifier (hex)
    -e, --end             End scan at this identifier (hex)\n
    additional required arguments for WINDOWS or Python 2:
    -i, --interface       python-can interface for the scan.
                          Depends on used interpreter and system,
                          see examples below. Any python-can interface can
                          be provided. Please see:
                          https://python-can.readthedocs.io for
                          further interface examples.
    optional arguments:
    -a, --python-can_args Additional arguments for a python-can Bus object.\n
    -h, --help            show this help message and exit
    -n NOISE_LISTEN_TIME, --noise_listen_time NOISE_LISTEN_TIME
                          Seconds listening for noise before scan.
    -t SNIFF_TIME, --sniff_time SNIFF_TIME
                          Duration in milliseconds a sniff is waiting for a
                          flow-control response.
    -x, --extended        Scan with ISOTP extended addressing.
                          This has nothing to do with extended CAN identifiers
    -C, --piso            Print 'Copy&Paste'-ready ISOTPSockets.
    -v, --verbose         Display information during scan.\n
        --extended_can_id Use extended CAN identifiers
    Example of use:\n
    Python2 or Windows:
        python2 -m scapy.tools.automotive.isotpscanner --interface=pcan --channel=PCAN_USBBUS1 --start 0 --end 100
    python2 -m scapy.tools.automotive.isotpscanner --interface=pcan --channel=PCAN_USBBUS1 -a 'bitrate=500000 fd=True' --start 0 --end 100
    python2 -m scapy.tools.automotive.isotpscanner --interface vector --channel 0 --start 0 --end 100
    python2 -m scapy.tools.automotive.isotpscanner --interface vector --channel 0 --python-can_args 'bitrate=500000, poll_interval=1' --start 0 --end 100
    python2 -m scapy.tools.automotive.isotpscanner --interface socketcan --channel=can0 --start 0 --end 100\n
    Python3 on Linux:
    python3 -m scapy.tools.automotive.isotpscanner --channel can0 --start 0 --end 100 \n''',  # noqa: E501
          file=sys.stderr if is_error else sys.stdout)


def create_socket(python_can_args, interface, channel):
    # type: (Optional[str], Optional[str], str) -> Tuple[CANSocket, str]

    if PYTHON_CAN:
        if python_can_args:
            interface_string = "CANSocket(bustype=" \
                               "'%s', channel='%s', %s)" % \
                               (interface, channel, python_can_args)
            arg_dict = dict((k, literal_eval(v)) for k, v in
                            (pair.split('=') for pair in
                             re.split(', | |,', python_can_args)))
            sock = CANSocket(bustype=interface, channel=channel,
                             **arg_dict)
        else:
            interface_string = "CANSocket(bustype=" \
                               "'%s', channel='%s')" % \
                               (interface, channel)
            sock = CANSocket(bustype=interface, channel=channel)
    else:
        sock = CANSocket(channel=channel)
        interface_string = "\"%s\"" % channel

    return sock, interface_string


def main():
    # type: () -> None
    extended = False
    piso = False
    verbose = False
    extended_can_id = False
    sniff_time = 100
    noise_listen_time = 2
    start = None
    end = None
    channel = None
    interface = None
    python_can_args = None
    conf.verb = -1

    options = getopt.getopt(
        sys.argv[1:],
        'vxCt:n:i:c:a:s:e:h:w',
        ['verbose', 'noise_listen_time=', 'sniff_time=', 'interface=', 'piso',
         'channel=', 'python-can_args=', 'start=', 'end=', 'help', 'extended',
         'extended_can_id'])

    try:
        for opt, arg in options[0]:
            if opt in ('-v', '--verbose'):
                verbose = True
            elif opt in ('-x', '--extended'):
                extended = True
            elif opt in ('-C', '--piso'):
                piso = True
            elif opt in ('-h', '--help'):
                usage(False)
                sys.exit(0)
            elif opt in ('-t', '--sniff_time'):
                sniff_time = int(arg)
            elif opt in ('-n', '--noise_listen_time'):
                noise_listen_time = int(arg)
            elif opt in ('-i', '--interface'):
                interface = arg
            elif opt in ('-c', '--channel'):
                channel = arg
            elif opt in ('-a', '--python-can_args'):
                python_can_args = arg
            elif opt in ('-s', '--start'):
                start = int(arg, 16)
            elif opt in ('-e', '--end'):
                end = int(arg, 16)
            elif opt in '--extended_can_id':
                extended_can_id = True
    except getopt.GetoptError as msg:
        usage(True)
        print("ERROR:", msg, file=sys.stderr)
        raise SystemExit

    if start is None or \
            end is None or \
            channel is None or \
            (PYTHON_CAN and interface is None):
        usage(True)
        print("\nPlease provide all required arguments.\n", file=sys.stderr)
        sys.exit(1)

    if end >= 2**29 or start >= 2**29:
        print("Argument 'start' and 'end' must be < " + hex(2**29),
              file=sys.stderr)
        sys.exit(1)

    if not extended_can_id and (end >= 0x800 or start >= 0x800):
        print("Standard can identifiers must be < 0x800.\n"
              "Use --extended_can_id option to scan with "
              "extended CAN identifiers.",
              file=sys.stderr)
        sys.exit(1)

    if end < start:
        print("start must be equal or smaller than end.", file=sys.stderr)
        sys.exit(1)

    try:
        sock, interface_string = \
            create_socket(python_can_args, interface, channel)

        if verbose:
            print("Start scan (%s - %s)" % (hex(start), hex(end)))

        stop_event = threading.Event()

        def signal_handler(*args):
            # type: (Any) -> None
            print('Interrupting scan!')
            stop_event.set()

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        result = isotp_scan(sock,
                            range(start, end + 1),
                            extended_addressing=extended,
                            noise_listen_time=noise_listen_time,
                            sniff_time=float(sniff_time) / 1000,
                            output_format="code" if piso else "text",
                            can_interface=interface_string,
                            extended_can_id=extended_can_id,
                            verbose=verbose,
                            stop_event=stop_event)

        print("Scan: \n%s" % result)

    except Exception as e:
        usage(True)
        print("\nSocket couldn't be created. Check your arguments.\n",
              file=sys.stderr)
        print(e, file=sys.stderr)
        sys.exit(1)

    finally:
        if sock is not None and not sock.closed:
            sock.close()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Andreas Korb <andreas.d.korb@gmail.com>
# Copyright (C) Nils Weiss <nils@we155.de>

# scapy.contrib.description = On Board Diagnostic Protocol (OBD-II)
# scapy.contrib.status = loads

import struct

from scapy.contrib.automotive.obd.iid.iids import *
from scapy.contrib.automotive.obd.mid.mids import *
from scapy.contrib.automotive.obd.pid.pids import *
from scapy.contrib.automotive.obd.tid.tids import *
from scapy.contrib.automotive.obd.services import *
from scapy.packet import bind_layers, NoPayload
from scapy.config import conf
from scapy.fields import XByteEnumField
from scapy.contrib.isotp import ISOTP

try:
    if conf.contribs['OBD']['treat-response-pending-as-answer']:
        pass
except KeyError:
    # log_automotive.info("Specify \"conf.contribs['OBD'] = "
    #                     "{'treat-response-pending-as-answer': True}\" to treat "
    #                     "a negative response 'requestCorrectlyReceived-"
    #                     "ResponsePending' as answer of a request. \n"
    #                     "The default value is False.")
    conf.contribs['OBD'] = {'treat-response-pending-as-answer': False}


class OBD(ISOTP):
    services = {
        0x01: 'CurrentPowertrainDiagnosticDataRequest',
        0x02: 'PowertrainFreezeFrameDataRequest',
        0x03: 'EmissionRelatedDiagnosticTroubleCodesRequest',
        0x04: 'ClearResetDiagnosticTroubleCodesRequest',
        0x05: 'OxygenSensorMonitoringTestResultsRequest',
        0x06: 'OnBoardMonitoringTestResultsRequest',
        0x07: 'PendingEmissionRelatedDiagnosticTroubleCodesRequest',
        0x08: 'ControlOperationRequest',
        0x09: 'VehicleInformationRequest',
        0x0A: 'PermanentDiagnosticTroubleCodesRequest',
        0x41: 'CurrentPowertrainDiagnosticDataResponse',
        0x42: 'PowertrainFreezeFrameDataResponse',
        0x43: 'EmissionRelatedDiagnosticTroubleCodesResponse',
        0x44: 'ClearResetDiagnosticTroubleCodesResponse',
        0x45: 'OxygenSensorMonitoringTestResultsResponse',
        0x46: 'OnBoardMonitoringTestResultsResponse',
        0x47: 'PendingEmissionRelatedDiagnosticTroubleCodesResponse',
        0x48: 'ControlOperationResponse',
        0x49: 'VehicleInformationResponse',
        0x4A: 'PermanentDiagnosticTroubleCodesResponse',
        0x7f: 'NegativeResponse'}

    name = "On-board diagnostics"

    fields_desc = [
        XByteEnumField('service', 0, services)
    ]

    def hashret(self):
        if self.service == 0x7f:
            return struct.pack('B', self.request_service_id & ~0x40)
        return struct.pack('B', self.service & ~0x40)

    def answers(self, other):
        if other.__class__ != self.__class__:
            return False
        if self.service == 0x7f:
            return self.payload.answers(other)
        if self.service == (other.service + 0x40):
            if isinstance(self.payload, NoPayload) or \
                    isinstance(other.payload, NoPayload):
                return True
            else:
                return self.payload.answers(other.payload)
        return False


# Service Bindings

bind_layers(OBD, OBD_S01, service=0x01)
bind_layers(OBD, OBD_S02, service=0x02)
bind_layers(OBD, OBD_S03, service=0x03)
bind_layers(OBD, OBD_S04, service=0x04)
bind_layers(OBD, OBD_S06, service=0x06)
bind_layers(OBD, OBD_S07, service=0x07)
bind_layers(OBD, OBD_S08, service=0x08)
bind_layers(OBD, OBD_S09, service=0x09)
bind_layers(OBD, OBD_S0A, service=0x0A)

bind_layers(OBD, OBD_S01_PR, service=0x41)
bind_layers(OBD, OBD_S02_PR, service=0x42)
bind_layers(OBD, OBD_S03_PR, service=0x43)
bind_layers(OBD, OBD_S04_PR, service=0x44)
bind_layers(OBD, OBD_S06_PR, service=0x46)
bind_layers(OBD, OBD_S07_PR, service=0x47)
bind_layers(OBD, OBD_S08_PR, service=0x48)
bind_layers(OBD, OBD_S09_PR, service=0x49)
bind_layers(OBD, OBD_S0A_PR, service=0x4A)
bind_layers(OBD, OBD_NR, service=0x7F)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
LLMNR (Link Local Multicast Node Resolution).

[RFC 4795]

LLMNR is based on the DNS packet format (RFC1035 Section 4)
RFC also envisions LLMNR over TCP. Like vista, we don't support it -- arno
"""

import struct

from scapy.fields import (
    BitEnumField,
    BitField,
    DestField,
    DestIP6Field,
    ShortField,
)
from scapy.packet import Packet, bind_layers, bind_bottom_up
from scapy.compat import orb
from scapy.layers.inet import UDP
from scapy.layers.dns import (
    DNSCompressedPacket,
    DNS_am,
    DNS,
    DNSQR,
    DNSRR,
)


_LLMNR_IPv6_mcast_Addr = "FF02:0:0:0:0:0:1:3"
_LLMNR_IPv4_mcast_addr = "224.0.0.252"


class LLMNRQuery(DNSCompressedPacket):
    name = "Link Local Multicast Node Resolution - Query"
    qd = []
    fields_desc = [
        ShortField("id", 0),
        BitField("qr", 0, 1),
        BitEnumField("opcode", 0, 4, {0: "QUERY"}),
        BitField("c", 0, 1),
        BitField("tc", 0, 1),
        BitField("t", 0, 1),
        BitField("z", 0, 4)
    ] + DNS.fields_desc[-9:]
    overload_fields = {UDP: {"sport": 5355, "dport": 5355}}

    def get_full(self):
        # Required for DNSCompressedPacket
        return self.original

    def hashret(self):
        return struct.pack("!H", self.id)

    def mysummary(self):
        s = self.__class__.__name__
        if self.qr:
            if self.an and isinstance(self.an[0], DNSRR):
                s += " '%s' is at '%s'" % (
                    self.an[0].rrname.decode(errors="backslashreplace"),
                    self.an[0].rdata,
                )
            else:
                s += " [malformed]"
        elif self.qd and isinstance(self.qd[0], DNSQR):
            s += " who has '%s'" % (
                self.qd[0].qname.decode(errors="backslashreplace"),
            )
        else:
            s += " [malformed]"
        return s, [UDP]


class LLMNRResponse(LLMNRQuery):
    name = "Link Local Multicast Node Resolution - Response"
    qr = 1

    def answers(self, other):
        return (isinstance(other, LLMNRQuery) and
                self.id == other.id and
                self.qr == 1 and
                other.qr == 0)


class _LLMNR(Packet):
    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if len(_pkt) >= 2:
            if (orb(_pkt[2]) & 0x80):  # Response
                return LLMNRResponse
            else:                  # Query
                return LLMNRQuery
        return cls


bind_bottom_up(UDP, _LLMNR, dport=5355)
bind_bottom_up(UDP, _LLMNR, sport=5355)
bind_layers(UDP, _LLMNR, sport=5355, dport=5355)

DestField.bind_addr(LLMNRQuery, _LLMNR_IPv4_mcast_addr, dport=5355)
DestField.bind_addr(LLMNRResponse, _LLMNR_IPv4_mcast_addr, dport=5355)
DestIP6Field.bind_addr(LLMNRQuery, _LLMNR_IPv6_mcast_Addr, dport=5355)
DestIP6Field.bind_addr(LLMNRResponse, _LLMNR_IPv6_mcast_Addr, dport=5355)


class LLMNR_am(DNS_am):
    """
    LLMNR answering machine.

    This has the same arguments as DNS_am. See help(DNS_am)

    Example::

        >>> llmnrd(joker="192.168.0.2", iface="eth0")
        >>> llmnrd(match={"TEST": "192.168.0.2"})
    """
    function_name = "llmnrd"
    filter = "udp port 5355"
    cls = LLMNRQuery


# LLMNRQuery(id=RandShort(), qd=DNSQR(qname="vista.")))
